import gustaf as gus

import splinepy

if __name__ == "__main__":
    # create 2p2d nurbs
    s_nurbs = splinepy.Bezier(
        degrees=[2, 1],
        control_points=[
            [0, 0],
            [0.5, -0.5],
            [1, 0],
            [0, 1],
            [0.5, 0.5],
            [1, 1],
        ],
    ).nurbs
    s_nurbs.insert_knots(0, [0.5])

    # extract line by specifying ranges
    line = s_nurbs.extract.spline({0: [0.4, 0.8], 1: 0.7})

    gus.show(
        ["Source spline", s_nurbs],
        ["extract.spline({0: [0.4, 0.8], 1: 0.7})", line, s_nurbs],
        ["Extracted spline", line],
    )



"""
io utils.
"""

import os as _os


def make_meaningful(line, comment="#"):
    """
    Tries to turn a str input a meaningful line. Returns False if it is not
    possible

    Parameters
    -----------
    line: str
    comment: str
      str that defined comment section. "#" is default

    Returns
    --------

    """
    line = line.strip()
    if len(line) == 0 or line.startswith(comment):
        return False
    else:
        return line


def next_line(f, comment="#"):
    """
    Strips and returns next meaningful line from the opened, readable file.
    It tries 513 times before it says it is the end. If it is end, it
    will return None.
    This is to prevent never ending while loop.

    Parameters
    -----------
    f: _io.TextIoWrapper
    comment: str
      A symbol that defines comment section. "#" is default.

    Returns
    --------
    next_meaningful_str: str
    """
    line = f.readline().strip()
    counter = 0
    while line.startswith(comment) or len(line) == 0:
        counter += 1
        if counter >= 513:
            return None
        line = f.readline().strip()

    return line


def form_lines(*args):
    """
    Formulate a string, taking each args as a line.

    Parameters
    -----------
    *args: str

    Returns
    --------
    line_separated_str: str
    """
    line_separated_str = ""
    for a in args:
        line_separated_str += a + "\n"

    return line_separated_str


def abs_fname(fname):
    """
    Checks if fname is absolute. If not, returns abspath. Tilde safe.

    Parameters
    ----------
    fname: str

    Returns
    --------
    abs_fname: str
      Maybe same to fname, maybe not.
    """
    if _os.path.isabs(fname):
        pass

    elif "~" in fname:
        fname = _os.path.expanduser(fname)

    else:
        fname = _os.path.abspath(fname)

    return fname


def expand_tabs(fname, overwrite=True, tab_expand=2):
    """Replaces tabs in a text file with spaces

    Parameters
    ----------
    fname : string
      Filename
    overwrite : bool
      Inplace modification, otherwise new file with prefix copy is created
    tab_expand : int
      Determines how many spaces to be set for each tab

    Returns
    -------
    None
    """
    import os.path as _op

    # Safe guard in case an absolute path is handed to the function
    if overwrite:
        out_name = fname
    else:
        dir, file = _op.split(fname)
        out_name = dir + "copy_" + file

    with open(fname) as inputFile:
        file_contents = inputFile.read()
    file_contents.replace("\t", " " * tab_expand)
    with open(out_name, "w") as exportFile:
        exportFile.write(file_contents)


def dict_to_spline(spline_dictionary):
    """Create a list of splines from a list of dictionaries

    Parameters
    ----------
    spline_dictionary : list
      List of dictionaries

    Returns
    -------
    spline_lits : list
      List of splines in called format (NAME_TO_TYPE)
    """
    from splinepy.settings import NAME_TO_TYPE as _NAME_TO_TYPE
    from splinepy.spline import Spline as _Spline

    spline_list = [_Spline(**spd) for spd in spline_dictionary]
    return [_NAME_TO_TYPE[spl.name](spline=spl) for spl in spline_list]



import xml.etree.ElementTree as _ET

import numpy as _np

try:
    from vedo import Plotter as _Plotter
    from vedo import color_map as _color_map
    from vedo.colors import get_color as _get_color
except ImportError as err:
    from gustaf.helpers.raise_if import ModuleImportRaiser

    _error_message_vedo_import = "Vedo library required for svg export"
    _Plotter = ModuleImportRaiser(_error_message_vedo_import, err)
    _color_map = ModuleImportRaiser(_error_message_vedo_import, err)
    _get_color = ModuleImportRaiser(_error_message_vedo_import, err)

from gustaf import Vertices as _Vertices

from splinepy.utils.log import debug as _debug
from splinepy.utils.log import warning as _warning


def _rgb_2_hex(r, g, b):
    """
    Helper function to convert (r,g,b) to hex format

    Parameters
    ----------
    r : float [0,1]
      red value
    g : float [0,1]
      green value
    b : float [0,1]
      blue value

    Returns
    -------
    hex : string
    """
    return f"#{int(255 * r):02x}{int(255 * g):02x}{int(255 * b):02x}"


def _write_png(bitmap, as_base64=True):
    """
    Write PNG data from a buffer.

    This function is taken from
    https://stackoverflow.com/questions/56564977/

    Nitesh Menon's answer with some comments added to it

    Replacing this by an external library like PIL or imageio is a @todo
    still.

    Parameters
    ----------
    bitmap : np.ndarray
        Picture with dimensions (width, height, 4) as RGBA image
    as_base64 : bool
        Write into readable base64 format

    Returns
    -------
    png : string
        Base64 encoded String with png data
    """
    import base64
    import struct  # Packing data
    import zlib  # Compression

    # Extract data
    buf = bytearray(bitmap)
    if (
        (bitmap.ndim != 3)
        or (bitmap.shape[2] != 4)
        or (bitmap.dtype is not _np.dtype("uint8"))
    ):
        raise ValueError("Was expecting bitmap in RGBA format")
    width = bitmap.shape[1]
    height = bitmap.shape[0]

    # Reverse the vertical line order and add null bytes at the start
    width_byte_4 = width * 4

    # Note: Contrary to the original post, the array is not concatenated in
    # reverse but in the original order (modified by jzwar)
    raw_data = b"".join(
        # Add a null byte and concatenate the raw data
        b"\x00" + buf[span : (span + width_byte_4)]
        # Iterate over the buffer
        for span in range(0, (height - 1) * width_byte_4 + 1, width_byte_4)
    )

    def png_pack(png_tag, data):
        """
        Pack PNG data into a chunk according to png standard.

        Parameters:
        png_tag : bytes
            PNG tag identifier.
        data : bytes
            Data to be packed.

        Returns
        -------
        bytes
            Packed PNG chunk.
        """
        chunk_head = png_tag + data  # Concatenate the tag and data
        return (
            # Pack the length of chunk
            struct.pack("!I", len(data))
            +
            # Concatenate the chunk header
            chunk_head
            +
            # Pack CRC-32 checksum
            struct.pack("!I", 0xFFFFFFFF & zlib.crc32(chunk_head))
        )

    pure_data = b"".join(
        [
            b"\x89PNG\r\n\x1a\n",  # PNG file signature
            png_pack(
                b"IHDR", struct.pack("!2I5B", width, height, 8, 6, 0, 0, 0)
            ),  # PNG header chunk
            # PNG data chunk (compressed)
            png_pack(b"IDAT", zlib.compress(raw_data, 9)),
            png_pack(b"IEND", b""),  # PNG end chunk
        ]
    )

    # Return encoded
    if as_base64:
        return base64.b64encode(pure_data)
    else:
        return pure_data


def _export_spline_field(spline, svg_element, box_min_x, box_max_y, **kwargs):
    """
    Export a spline's data-field as a pixel-bitmap. The pixel density is based
    on the sample resolution, (10 * resolution). The image is embedded in b64
    format

    Parameters
    ----------
    spline : Spline
      Providing the control points for the mesh
    svg_spline_element : xml.etree.ElementTree
      Parent element of the spline (most likely spline group), where the mesh
      is written into
    box_min_x : float/int
      minimum coordinate of the box, providing the necessary offsets for spline
      export
    box_max_y : float/int
      maximum y coordinate of the surrounding box (pictures use LHS coordinate
      system in top left corner)

    Returns
    -------
    None
    """

    from splinepy.helpme.visualize import _process_scalar_field, _sample_spline

    # Sample the spline
    resolution = spline.show_options.get("resolutions", 100)
    sampled_spline = _sample_spline(spline, resolution)
    data = spline.show_options.get("data", None)

    # Check
    if data is None:
        raise ValueError(
            "There is no data provided, although data plot is requested"
        )

    # Process the scalar field
    _process_scalar_field(spline, data, sampled_spline, res=resolution)
    # scalar bar distorts the dimensions of spline bitmap image
    sampled_spline.show_options["scalarbar"] = False
    # Set a resolution (This needs to be improved still)
    pixel_size = (resolution * 10, resolution * 10)

    # Create a Plotter
    plotter = _Plotter(
        shape=(1, 1),  # Only one field
        N=1,  # Number of things to plot
        size=pixel_size,  # Resolution, i.e. pixel density
        sharecam=True,
        offscreen=True,
        title="",
        bg=(255, 255, 255),
        axes=0,
    )
    showable = sampled_spline.showable()

    # Discretize colors
    if kwargs.get("n_colors", None) is not None:
        n_colors = kwargs.get("n_colors", None)
        cmap_key = spline.show_options.get("cmap", "jet")
        vmin = kwargs.get("vmin", spline.show_options.get("vmin", None))
        vmax = kwargs.get("vmax", spline.show_options.get("vmax", None))
        showable.cmap(cmap_key, vmin=vmin, vmax=vmax, n_colors=n_colors)

    plotter.show(showable, zoom="tightest")

    # Extract bounding box
    x_min, y_min, x_max, y_max = sampled_spline.bounds().ravel()

    # Extract the bitmap and transform to RGBA
    bitmap = plotter.screenshot(asarray=True)

    if bitmap.shape[2] == 3:
        alpha_layer = (
            _np.ones((bitmap.shape[0], bitmap.shape[1]), dtype=bitmap.dtype)
            * 255
        )
        alpha_layer[_np.all(bitmap == 255, axis=-1)] = 0
        bitmap = _np.concatenate(
            (bitmap, alpha_layer.reshape(*alpha_layer.shape, 1)), axis=-1
        )
    elif bitmap.shape[2] == 4:
        alpha_layer = bitmap[:, :, -1]
    else:
        raise RuntimeError("Failed to extract field bitmap")

    # Crop image
    pixel_layer = alpha_layer != 0
    has_pixels = _np.where(_np.any(pixel_layer, axis=1))[0]
    min_x_pixel = has_pixels.min()
    max_x_pixel = has_pixels.max()
    has_pixels = _np.where(_np.any(pixel_layer, axis=0))[0]
    min_y_pixel = has_pixels.min()
    max_y_pixel = has_pixels.max()
    bitmap = bitmap[min_x_pixel:max_x_pixel, min_y_pixel:max_y_pixel, :]

    # Write Bitmap into svg
    image_svg = _ET.SubElement(
        svg_element,
        "image",
        x=str(x_min - box_min_x),  # str(box_max_y - xy)
        y=str(box_max_y - y_max),
        width=str(x_max - x_min),
        height=str(y_max - y_min),
    )
    image_svg.attrib["xlink:href"] = "data:image/png;base64," + _write_png(
        bitmap=bitmap, as_base64=True
    ).decode("utf-8")


def _export_gustaf_object(
    gus_object, svg_spline_element, box_min_x, box_max_y, **kwargs
):
    """
    Exports gustaf-type objects in svg format

    Parameters
    """
    r, g, b = _get_color(gus_object.show_options.get("c", "red"))
    a = gus_object.show_options.get("alpha", 1.0)
    radius = gus_object.show_options.get("r", 0.1)

    if isinstance(gus_object, _Vertices):
        data_name = gus_object.show_options.get("data", None)
        if data_name is not None:
            # Retrieve information on colors and values
            values = gus_object.vertex_data[data_name]
            v_min = gus_object.show_options.get("vmin", _np.min(values))
            v_max = gus_object.show_options.get("vmax", _np.max(values))
            cmap_style = gus_object.show_options.get("cmap", "jet")
            colors = [
                _rgb_2_hex(r, g, b)
                for r, g, b, in _color_map(
                    values.ravel(), name=cmap_style, vmin=v_min, vmax=v_max
                )
            ]
        else:
            colors = [_rgb_2_hex(r, g, b)] * gus_object.vertices.shape[0]

        svg_control_points = _ET.SubElement(
            svg_spline_element,
            "g",
            id="control_points",
        )

        for vertex, color in zip(gus_object.vertices, colors):
            _ET.SubElement(
                svg_control_points,
                "circle",
                cx=str(vertex[0] - box_min_x),
                cy=str(box_max_y - vertex[1]),
                r=str(radius),
                style=f"fill:{color};fill-opacity:{a}",
            )

        labels = gus_object.show_options.get("labels", None)
        if (labels is None) and (
            gus_object.show_options.get("vertex_ids", None) is not None
        ):
            labels = [str(i) for i in range(gus_object.vertices.shape[0])]
        if labels is not None:
            svg_labels = _ET.SubElement(
                svg_spline_element,
                "g",
                id="labels",
            )

            # Set text options
            if kwargs.get("font_family", None) is not None:
                svg_labels.attrib["font-family"] = kwargs["font_family"]

            svg_labels.attrib["font-size"] = str(kwargs.get("font_size", 0.1))
            svg_labels.attrib["text-anchor"] = kwargs.get(
                "text_anchor", "middle"
            )
            svg_labels.attrib["fill"] = _rgb_2_hex(
                *_get_color(kwargs.get("text_color", "k"))
            )
            svg_labels.attrib["stroke"] = "none"
            dx = radius

            for ctp, label in zip(gus_object.vertices, labels):
                text_element = _ET.SubElement(
                    svg_labels,
                    "text",
                    x=str(ctp[0] - box_min_x),
                    y=str(box_max_y - ctp[1]),
                    dx=str(dx),
                    dy=str(dx),
                )
                text_element.text = label

    else:
        raise NotImplementedError(
            "Sorry, this feature is still being worked on. So far, svg_export "
            "is limited to vertices"
        )


def _export_control_mesh(
    spline, svg_spline_element, box_min_x, box_max_y, **kwargs
):
    """
    Export a spline's control mesh in svg format using polylines for the mesh
    lines, circles for the control points.

    Parameters
    ----------
    spline : Spline
      Providing the control points for the mesh
    svg_spline_element : xml.etree.ElementTree
      Parent element of the spline (most likely spline group), where the mesh
      is written into
    box_min_x : float/int
      minimum coordinate of the box, providing the necessary offsets for spline
      export
    box_max_y : float/int
      maximum y coordinate of the surrounding box (pictures use LHS coordinate
      system in top left corner)
    kwargs :
      Key word argument options for output specification
      `{"font-family","font-size","text-anchor","font-family","fill","stroke"}`

    Returns
    -------
    None
    """

    # Default behaviour in show is, no mesh and no ids, if no control points
    control_point_requested = spline.show_options.get("control_points", True)

    svg_mesh = _ET.SubElement(
        svg_spline_element,
        "g",
        id="control_mesh",
    )

    # First draw mesh
    if spline.show_options.get("control_mesh", control_point_requested):
        # Relevant options:
        # - control_mesh_c
        # - control_mesh_lw
        # - control_mesh_alpha (tbd)
        r, g, b = _get_color(spline.show_options.get("control_mesh_c", "red"))
        a = spline.show_options.get("control_mesh_alpha", 1.0)
        stroke_width = spline.show_options.get(
            "control_mesh_lw",
            _np.linalg.norm(_np.diff(spline.control_point_bounds)) * 0.01,
        )

        # Create a new group
        svg_mesh_polylines = _ET.SubElement(
            svg_mesh,
            "g",
            id="mesh",
            style=(
                f"fill:none;stroke:{_rgb_2_hex(r,g,b)};stroke-opacity:{a};"
                f"stroke-width:{stroke_width};stroke-linecap:round"
            ),
        )

        # Draw the actual lines
        if spline.para_dim == 1:
            _ET.SubElement(
                svg_mesh_polylines,
                "polyline",
                points=" ".join(
                    [
                        str(xx - box_min_x) + "," + str(box_max_y - xy)
                        for (xx, xy) in spline.cps
                    ]
                ),
            )
        if spline.para_dim == 2:
            mi = spline.multi_index
            cmr_xi, cmr_eta = spline.control_mesh_resolutions
            for i in range(cmr_eta):
                # for
                _ET.SubElement(
                    svg_mesh_polylines,
                    "polyline",
                    points=" ".join(
                        [
                            str(xx - box_min_x) + "," + str(box_max_y - xy)
                            for (xx, xy) in spline.cps[mi[:, i]]
                        ]
                    ),
                )
            for i in range(cmr_xi):
                # for
                _ET.SubElement(
                    svg_mesh_polylines,
                    "polyline",
                    points=" ".join(
                        [
                            str(xx - box_min_x) + "," + str(box_max_y - xy)
                            for (xx, xy) in spline.cps[mi[i, :]]
                        ]
                    ),
                )

    # Then control points
    if control_point_requested:

        # Relevant options:
        # - control_point_ids
        # - control_point_alpha
        # - control_point_c
        # - control_point_r

        # transform color into (relative) rgb using vedo helper function
        r, g, b = _get_color(spline.show_options.get("control_point_c", "red"))
        a = spline.show_options.get("control_point_alpha", 1.0)

        svg_control_points = _ET.SubElement(
            svg_mesh,
            "g",
            id="control_points",
            style=f"fill:{_rgb_2_hex(r, g, b)};fill-opacity:{a}",
        )

        for ctp in spline.control_points:
            _ET.SubElement(
                svg_control_points,
                "circle",
                cx=str(ctp[0] - box_min_x),
                cy=str(box_max_y - ctp[1]),
                r=str(spline.show_options.get("control_point_r", 0.02)),
            )

    # Lastly IDs
    if spline.show_options.get("control_point_ids", control_point_requested):
        # Text Options
        svg_control_point_ids = _ET.SubElement(
            svg_mesh,
            "g",
            id="control_point_ids",
        )

        # Set text options
        if kwargs.get("font_family", None) is not None:
            svg_control_point_ids.attrib["font-family"] = kwargs["font_family"]
        svg_control_point_ids.attrib["font-size"] = str(
            kwargs.get("font_size", 0.1)
        )
        svg_control_point_ids.attrib["text-anchor"] = kwargs.get(
            "text_anchor", "middle"
        )
        svg_control_point_ids.attrib["fill"] = _rgb_2_hex(
            *_get_color(kwargs.get("text_color", "k"))
        )
        svg_control_point_ids.attrib["stroke"] = "none"

        # Text offset
        dx = spline.show_options.get("control_point_r", 0.02) * 2

        for i, ctp in enumerate(spline.control_points):
            text_element = _ET.SubElement(
                svg_control_point_ids,
                "text",
                x=str(ctp[0] - box_min_x),
                y=str(box_max_y - ctp[1]),
                dx=str(dx),
                dy=str(dx),
            )
            text_element.text = str(i)


def _quiver_plot(
    spline, svg_spline_element, box_min_x, box_max_y, tolerance=0.0, **kwargs
):
    """
    Export a quiver plot using arrows as defined
    https://matplotlib.org/stable/_images/quiver_sizes.svg

    The arrows will be scaled using the `arrow_data_scale` flag

    Parameters
    ----------
    spline : Spline
      Providing the control points for the mesh
    svg_spline_element : xml.etree.ElementTree
      Parent element of the spline (most likely spline group), where the mesh
      is written into
    box_min_x : float/int
      minimum coordinate of the box, providing the necessary offsets for spline
      export
    box_max_y : float/int
      maximum y coordinate of the surrounding box (pictures use LHS coordinate
      system in top left corner)
    box_margins : float/int
      box margins are currently used to determine the text-offset, more options
      will be added soon
    tolerance : float
      tolerance as a cut-off length under which no arrow is plotted
    kwargs :
      Optional for arrow sizing,
      `{"arrow_width", "arrow_head_width", "arrow_head_length",
      "arrow_headaxis_length", "arrow_data_tolerance"}`

    Returns
    -------
    None
    """

    # Get arrow scales
    arrow_width = kwargs.get("arrow_width", 0.1)
    arrow_head_width = kwargs.get("arrow_head_width", 0.3)
    arrow_head_length = kwargs.get("arrow_head_length", 0.5)
    arrow_headaxis_length = kwargs.get("arrow_headaxis_length", 0.45)
    cut_off_tolerance = kwargs.get("arrow_data_tolerance", tolerance)

    # Create a default arrow along x-axis with length 1 that will be scaled
    # down accordingly
    default_arrow_control_points_ = _np.array(
        [
            [0, -arrow_width],
            [1 - arrow_headaxis_length, -arrow_width],
            [1 - arrow_head_length, -arrow_head_width],
            [1, 0],
            [1 - arrow_head_length, arrow_head_width],
            [1 - arrow_headaxis_length, arrow_width],
            [0, arrow_width],
        ]
    ) * [1, 0.5]

    # Obtain Data to be plotted
    arrow_data_field = spline.show_options.get("arrow_data", None)
    if arrow_data_field is None:
        return None

    # Save original scale for set back
    original_scale = spline.show_options.get("arrow_data_scale", None)
    spline.show_options["arrow_data_scale"] = 1

    # Use Arrow Data to retrieve spline data (prevent from scaling)
    arrow_data = spline.extract.arrow_data(arrow_data_field)
    positions = arrow_data.vertices[arrow_data.edges[:, 0]]
    directions = arrow_data.vertices[arrow_data.edges[:, 1]] - positions

    # Set back data to original options
    if original_scale is None:
        spline.show_options.pop("arrow_data_scale")
    else:
        spline.show_options["arrow_data_scale"] = original_scale

    # Retrieve scaling
    arrow_data_scale = spline.show_options.get("arrow_data_scale", 1)

    # Discard all points that are below the tolerance
    arrow_length = _np.linalg.norm(directions, axis=1)
    over_tolerance = arrow_length > cut_off_tolerance
    directions = directions[over_tolerance]
    positions = positions[over_tolerance]
    arrow_length = arrow_length[over_tolerance]

    # Create rotation matrices
    angles = _np.arctan2(directions[:, 1], directions[:, 0])
    rotation_matrices = _np.empty((angles.shape[0], 2, 2))
    rotation_matrices[:, 0, 0] = _np.cos(angles)
    rotation_matrices[:, 1, 0] = _np.sin(angles)
    rotation_matrices[:, 0, 1] = -rotation_matrices[:, 1, 0]
    rotation_matrices[:, 1, 1] = rotation_matrices[:, 0, 0]

    # Retrieve information on colors and values
    arrow_color = spline.show_options.get("arrow_data_color", None)
    if arrow_color is not None:
        colors = [_get_color(arrow_color)] * arrow_length.size
    else:
        v_min = spline.show_options.get("vmin", _np.min(arrow_length))
        v_max = spline.show_options.get("vmax", _np.max(arrow_length))
        cmap_style = spline.show_options.get("cmap", "jet")
        colors = _color_map(
            arrow_length,
            name=cmap_style,
            vmin=v_min,
            vmax=v_max,
        )

    # Map arrow control points
    arrow_control_points_ = _np.einsum(
        "nij,kj,n->nki",
        rotation_matrices,
        default_arrow_control_points_,
        arrow_length,
    )
    # Scaling
    arrow_control_points_ *= arrow_data_scale

    # Create a new group
    svg_quiver = _ET.SubElement(
        svg_spline_element,
        "g",
        id="quiver",
    )

    # Loop over arrows and write them into a group
    for i in range(arrow_length.shape[0]):
        _ET.SubElement(
            svg_quiver,
            "polyline",
            points=" ".join(
                [
                    str(xx - box_min_x) + "," + str(box_max_y - xy)
                    for (xx, xy) in (
                        arrow_control_points_[i, :, :] + positions[i, :]
                    )
                ]
            ),
            style=(f"fill:{_rgb_2_hex(*colors[i])};stroke:none"),
        )

    return None


def _add_scalar_bar(svg_element, box_size, **kwargs):
    """
    All sizes passed as kwargs are in relation to the screen window

    Parameters
    ----------
    svg_element : ElementTree
      Parent Element in the svg file
    box_size : tuple<float>
      Total size (in relative coordinate system) of the  picture

    kwargs:
      Specify more output options:

      - background (color, string, rgb, None) : background color
      - n_ticks : (int) Number of Tick mars
      - vmin : (float) Minimum Value - Required for colorbar
      - vmax : (float) Maximum Value - Required for colorbar
      - cmap : (color map string) Defaults to "jet"
      - scalarbar_width : (float) Width of the colorbar relative to picture
      - scalarbar_aspect_ratio : (float) Aspect ratio of colorbar
      - scalarbar_offset : (float) reserved space below picture
      - scalarbar_stroke_width : (float) boundary width
      - scalarbar_tick_distance: (float)distance from tick marks to colorbar
      - scalarbar_font_size : (float)
    """
    # Prepare group
    svg_scalarbar = _ET.SubElement(
        svg_element,
        "g",
        id="scalarbar",
    )

    # Scalar bar consists of 3 components
    # 1. Bar
    # 2. Tickmarks
    # 3. Ticks

    # Tick marks
    number_of_ticks = kwargs.get("n_ticks", 3)
    vmin = kwargs["vmin"]
    vmax = kwargs["vmax"]
    cmap = kwargs.get("cmap", "jet")

    scalarbar_width = kwargs.get("scalarbar_width", 0.4)
    scalarbar_aspect_ratio = kwargs.get("scalarbar_aspect_ratio", 10)
    scalarbar_offset = kwargs["scalarbar_offset"]

    # Temporary
    r, g, b = _get_color(kwargs.get("scalarbar_c", "k"))
    a = 1.0
    stroke_width = kwargs.get("scalarbar_stroke_width", box_size[0] * 0.002)

    # Prepare auxiliary values
    scalarbar_height = scalarbar_width / scalarbar_aspect_ratio
    min_x = box_size[0] * (1 - scalarbar_width) / 2
    max_x = box_size[0] * (1 + scalarbar_width) / 2
    center_y = box_size[1] * (scalarbar_offset) / (1 + 2 * scalarbar_offset)
    min_y = center_y + box_size[0] * scalarbar_height / 2
    max_y = center_y - box_size[0] * scalarbar_height / 2

    # Create bar (Using existing framework)
    from splinepy.bezier import Bezier as _Bezier
    from splinepy.helpme.create import box as _Box

    box = _Box(max_x - min_x, max_y - min_y)
    box.cps[:] += [min_x, min_y]
    scalar_values = _Bezier([1, 0], [[vmin], [vmax]])
    box.spline_data["bar"] = scalar_values
    box.show_options["data"] = "bar"
    box.show_options["cmap"] = cmap
    box.show_options["vmin"] = vmin
    box.show_options["vmax"] = vmax
    _export_spline_field(box, svg_scalarbar, 0, box_size[1], **kwargs)

    # Add tickmarks
    svg_tick_marks = _ET.SubElement(
        svg_scalarbar,
        "g",
        id="tick_mark_lines",
        style=(
            f"fill:none;stroke:{_rgb_2_hex(r,g,b)};stroke-opacity:{a};"
            f"stroke-width:{stroke_width};stroke-linecap:round"
        ),
    )
    svg_tick_labels = _ET.SubElement(
        svg_scalarbar,
        "g",
        id="svg_tick_labels",
    )

    # Set text options
    if kwargs.get("font_family", None) is not None:
        svg_tick_labels.attrib["font-family"] = kwargs["font_family"]
    svg_tick_labels.attrib["font-size"] = str(
        kwargs.get("scalarbar_font_size", stroke_width * 20)
    )
    svg_tick_labels.attrib["text-anchor"] = kwargs.get("text_anchor", "middle")
    svg_tick_labels.attrib["fill"] = _rgb_2_hex(
        *_get_color(kwargs.get("text_color", "k"))
    )
    svg_tick_labels.attrib["stroke"] = "none"
    # Text offset
    dx = kwargs.get("scalarbar_tick_distance", stroke_width * 10)

    # Create Ticks
    tick_marks = _np.linspace(vmin, vmax, number_of_ticks)
    tick_mark_positions = _np.linspace(min_x, max_x, number_of_ticks)

    # Separate groups for ticks
    _ET.SubElement(
        svg_tick_marks,
        "polyline",
        points=" ".join(
            [
                str(xx) + "," + str(box_size[1] - xy)
                for (xx, xy) in box.cps[[0, 1, 3, 2]]
            ]
        ),
    )

    for mark, position in zip(tick_marks, tick_mark_positions):
        # Draw a line
        _ET.SubElement(
            svg_tick_marks,
            "line",
            x1=str(position),
            x2=str(position),
            y1=str(box_size[1] - max_y),
            y2=str(box_size[1] - max_y + 1.5 * (max_y - min_y)),
        )
        text_element = _ET.SubElement(
            svg_tick_labels,
            "text",
            x=str(position),
            y=str(box_size[1] - max_y + 1.5 * (max_y - min_y)),
            dx=str(0),
            dy=str(-dx),
        )
        text_element.text = str(mark)

    pass


def _export_spline(
    spline, svg_spline_element, box_min_x, box_max_y, tolerance=None, **kwargs
):
    """
    Export a spline in svg format cubic beziers as approximations (even for
    lower degree splines, for simplicity). Also exports the knot lines

    Parameters
    ----------
    spline : Spline
      Providing the control points for the mesh
    svg_spline_element : xml.etree.ElementTree
      Parent element of the spline (most likely spline group), where the mesh
      is written into
    box_min_x : float/int
      minimum coordinate of the box, providing the necessary offsets for spline
      export
    box_max_y : float/int
      maximum y coordinate of the surrounding box (pictures use LHS coordinate
      system in top left corner)
    box_margins : float/int
      box margins are currently used to determine the text-offset, more options
      will be added soon
    tolerance : float
      Splines that can not be represented exactly by cubic B-Splines (i.e.,
      rationals or high order splines) a tolerance is required for the
      approximation. Default uses an absolute deviation of 1% of the bounding
      box of the given spline
    kwargs:
      `{"linecap"}`

    Returns
    -------
    None
    """
    from splinepy.helpme.fit import curve as _fit_curve
    from splinepy.helpme.reparametrize import flip_axes as _flip_axes
    from splinepy.settings import TOLERANCE

    # Maximum number of refinements for approximation
    MAX_ITERATION = 4

    # Set tolerance for export to default if no user data
    if tolerance is None:
        tolerance = 0.01 * _np.linalg.norm(
            spline.control_point_bounds[0, :]
            - spline.control_point_bounds[1, :]
        )

    # Sanity check
    if spline.para_dim not in [1, 2]:
        raise ValueError("Unsupported spline dimension")

    # Approximation curve-wise
    def _approximate_curve(original_spline, tolerance):
        """
        Approximate a curve using third order B-Splines

        Continuity will be preserved

        original_spline : Spline
          Any type spline as a basis for approximation
        tolerance : double
          approximation tolerance (given a maximum number of iterations)

        Returns
        -------
        """
        if original_spline.degrees[0] <= 3 and not (
            original_spline.is_rational and original_spline.degrees[0] > 1
        ):
            spline_approximation = original_spline.copy()
            spline_approximation.elevate_degrees(
                [0] * (3 - original_spline.degrees[0])
            )
        else:
            # Use fit tool to approximate curve
            _debug(
                "SVG export only supports (up to) cubic polynomial splines --"
                " using approximation"
            )

            # Queries
            para_queries = original_spline.greville_abscissae(
                duplicate_tolerance=TOLERANCE
            )

            # For rational splines this might be insufficient
            if original_spline.is_rational and (
                original_spline.degrees[0] < 3
            ):
                para_queries = _np.sort(
                    _np.vstack(
                        (
                            para_queries,
                            _np.convolve(
                                para_queries.ravel(),
                                _np.ones(2) * 0.5,
                                "valid",
                            ).reshape(-1, 1),
                        )
                    ),
                    axis=0,
                )

            # Create knot-vector
            k_mult = original_spline.knot_multiplicities[0]
            k_mult[1:-1] = _np.maximum(
                1, 3 - original_spline.degrees[0] + k_mult[1:-1]
            )
            k_mult[0] = 4
            k_mult[-1] = 4
            new_knot_vector = _np.repeat(original_spline.unique_knots, k_mult)
            residual = 2 * tolerance
            spline_approximation, residual = _fit_curve(
                original_spline.evaluate(para_queries),
                degree=3,
                knot_vector=new_knot_vector,
                associated_queries=para_queries,
            )

            # Refine until approximation is found
            for _ in range(MAX_ITERATION):
                if residual < tolerance:
                    break
                # Loop until tolerance satisfied
                para_queries = _np.sort(
                    _np.vstack(
                        (
                            para_queries,
                            _np.convolve(
                                para_queries.ravel(),
                                _np.ones(2) * 0.5,
                                "valid",
                            ).reshape(-1, 1),
                        )
                    ),
                    axis=0,
                )

                # Insert a few knots
                spline_approximation.insert_knots(
                    0,
                    _np.convolve(
                        spline_approximation.unique_knots[0].ravel(),
                        _np.ones(2) * 0.5,
                        "valid",
                    ),
                )
                new_knot_vector = spline_approximation.knot_vectors[0]

                # Create better approximation
                spline_approximation, residual = _fit_curve(
                    original_spline.evaluate(para_queries),
                    degree=3,
                    knot_vector=new_knot_vector,
                    associated_queries=para_queries,
                )

            if residual > tolerance:
                _warning(
                    "Requested tolerance could not be reached within maximum"
                    " number of refinement steps"
                )

        # Sanity check
        if not (
            (spline_approximation.degrees[0] == 3)
            or (not spline_approximation.is_rational)
            or (
                original_spline.is_rational and original_spline.degrees[0] <= 1
            )
        ):
            raise RuntimeError(
                "Spline Approximation returned unexpected result that is "
                "non-compliant to svg standard. Expected cubic bezier spline."
            )

        return spline_approximation

    # Write the actual spline as the lowest layer
    svg_spline = _ET.SubElement(
        svg_spline_element,
        "g",
        id="spline_paths",
    )

    # Check if a field is to be plotted
    if spline.show_options.get("data", None) is not None:

        # spline.show()
        _export_spline_field(
            spline, svg_spline, box_min_x, box_max_y, **kwargs
        )

    else:
        # Export is done in 2 stages
        # (1) Approximation (trying to preserve the continuity)
        # (2) Export

        # First : Retrieve export options
        r, g, b = _get_color(spline.show_options.get("c", "green"))
        a = spline.show_options.get("alpha", 1.0)

        # check if the parametric dimension is 2, if the case, extract boundary
        if spline.para_dim == 1:
            # Approximation
            spline_copy = _approximate_curve(spline, tolerance)
            bezier_elements = spline_copy.extract.beziers()
            path_d = (
                f"M {bezier_elements[0].cps[0,0]-box_min_x},"
                f"{box_max_y - bezier_elements[0].cps[0,1]}"
            )
            path_d += " ".join(
                [
                    (
                        f" C {s.cps[1,0]-box_min_x},{box_max_y - s.cps[1,1]}"
                        f" {s.cps[2,0]-box_min_x},{box_max_y - s.cps[2,1]}"
                        f" {s.cps[3,0]-box_min_x},{box_max_y - s.cps[3,1]}"
                    )
                    for s in bezier_elements
                ]
            )
            _ET.SubElement(
                svg_spline,
                "path",
                # draw path
                d=path_d,
                style=(
                    f"fill:none;stroke:{_rgb_2_hex(r,g,b)};stroke-opacity:{a};"
                    f"stroke-width:{spline.show_options.get('lw', 0.01)};"
                    "stroke-linecap:round"
                ),
            )
        else:
            # Extract boundaries
            spline_boundaries = spline.extract.boundaries()

            # Flip boundaries on boundary 0 and 3
            _flip_axes(spline_boundaries[0], axes=[0], inplace=True)
            _flip_axes(spline_boundaries[3], axes=[0], inplace=True)

            bezier_elements = []
            for i in [2, 1, 3, 0]:
                spline_copy = _approximate_curve(
                    spline_boundaries[i], tolerance
                )
                bezier_elements += spline_copy.extract.beziers()

            path_d = (
                f"M {bezier_elements[0].cps[0,0]-box_min_x},"
                f"{box_max_y - bezier_elements[0].cps[0,1]}"
            )
            path_d += " ".join(
                [
                    (
                        f" C"
                        f" {s.cps[1,0]-box_min_x},{box_max_y - s.cps[1,1]}"
                        f" {s.cps[2,0]-box_min_x},{box_max_y - s.cps[2,1]}"
                        f" {s.cps[3,0]-box_min_x},{box_max_y - s.cps[3,1]}"
                    )
                    for s in bezier_elements
                ]
            )

            _ET.SubElement(
                svg_spline,
                "path",
                # draw path
                d=path_d,
                style=(
                    f"fill:{_rgb_2_hex(r,g,b)};fill-opacity:{a};stroke:none;"
                    f"stroke-linecap:{kwargs.get('linecap','round')}"
                ),
            )

        # Plot knots as rectangles
    if spline.show_options.get("knots", True):
        # Exports knots in separate group
        svg_knots = _ET.SubElement(
            svg_spline_element,
            "g",
            id="knots",
        )

        # Relevant options
        # - knot_alpha
        # - knot_c
        # - knot_lw
        r, g, b = _get_color(spline.show_options.get("knot_c", "blue"))
        a = spline.show_options.get("knot_alpha", 1.0)
        lw = spline.show_options.get("knot_lw", 0.02)
        if spline.para_dim == 1:
            ukvs = spline.unique_knots[0]
            knot_projected = spline.evaluate(ukvs.reshape(-1, 1))
            for x, y in knot_projected:

                _ET.SubElement(
                    svg_knots,
                    "rect",
                    x=str(x - box_min_x - 0.5 * lw),
                    y=str(box_max_y - y - 0.5 * lw),
                    height=str(lw),
                    width=str(lw),
                    style=(
                        f"fill:{_rgb_2_hex(r,g,b)};stroke:none;"
                        f"fill-opacity:{a};"
                    ),
                )

        else:

            # Extract knot lines as splines
            knot_lines = []
            for knot in spline.unique_knots[0]:
                knot_lines.append(spline.extract.spline(0, knot))
            for knot in spline.unique_knots[1]:
                knot_lines.append(spline.extract.spline(1, knot))

            # Export knots to paths
            for knot_line in knot_lines:
                # Approximation and export
                spline_copy = _approximate_curve(knot_line, tolerance)
                bezier_elements = spline_copy.extract.beziers()
                path_d = (
                    f"M {bezier_elements[0].cps[0,0]-box_min_x},"
                    f"{box_max_y - bezier_elements[0].cps[0,1]}"
                )
                path_d += " ".join(
                    [
                        (
                            f" C {s.cps[1,0]-box_min_x},"
                            f"{box_max_y - s.cps[1,1]}"
                            f" {s.cps[2,0]-box_min_x},{box_max_y - s.cps[2,1]}"
                            f" {s.cps[3,0]-box_min_x},{box_max_y - s.cps[3,1]}"
                        )
                        for s in bezier_elements
                    ]
                )
                _ET.SubElement(
                    svg_knots,
                    "path",
                    # draw path
                    d=path_d,
                    style=(
                        f"fill:none;stroke:{_rgb_2_hex(r,g,b)};"
                        f"stroke-opacity:{a};"
                        f"stroke-width:{lw};"
                        f"stroke-linecap:{kwargs.get('linecap','round')}"
                    ),
                )


def export(
    fname,
    *objects_to_plot,
    indent=True,
    box_margins=0.1,
    tolerance=None,
    **kwargs,
):
    """
    Exports a number of splines into an svg plot

    Graphics are built to resemble vedo plots as close as possible by relying
    on spline show_options

    Example
    -------
    .. code-block:: python

        spline = c.splinepy.Bezier(
            degrees=[2, 2],
            control_points=[
                [1, 0],
                [3, 1],
                [5, 0],  # First row
                [0, 2],
                [2, 3],
                [4, 2],  # Second row
                [1, 4],
                [3, 5],
                [5, 4],  # Third row
            ],
        )
        spline.show_options["c"] = (157, 157, 156)
        spline.show_options["knot_c"] = "red"
        spline.show_options["control_mesh"] = True
        spline.show_options["control_mesh_lw"] = 0.05
        spline.show_options["control_point_c"] = (0, 102, 157)
        spline.show_options["control_mesh_c"] = (0, 102, 157)
        spline.show_options["control_point_ids"] = False
        spline.show_options["control_point_r"] = 0.2
        splinepy.io.svg.export(
            "spline.svg",
            spline,
            box_margins=0.2,
            background=None,
        )

    results in

    .. raw:: html

        <object
            data="../_static/svg_mini_example.svg"
            type="image/svg+xml"
            align="center"
            >
        </object>


    Parameters
    ----------
    fname : string
        name of the output file
    splines : Spline-Type, list, Gustaf-type
        Splines to be exported
    indent: bool
      Appends white spaces using xml.etree.ElementTree.indent, if possible.
      Makes it more human-readable
    box_margins : int, float
      Offset around the splines (relative units)
    tolerance : float
      Splines that can not be represented exactly by cubic B-Splines (i.e.,
      rationals or high order splines) a tolerance is required for the
      approximation. Default uses an absolute deviation of 1% of the bounding
      box of the given spline
    kwargs:
      Specify more output options:

      - background (color, string, rgb, None) : background color
      - linecap ("string"): linecap option for end of lines and connections, see
        https://www.w3.org/TR/SVG2/
        for more information
      - font_family (string): Font for control point ids (optional)
      - font_size (float): Size of control point ids
      - text_anchor (string): position of the control point ids relative to
        point
      - text_color (color, string, rgb) : control point id color
      - arrow_data_tolerance : (float) Cutt off for arrows that are too small
      - arrow_width (float): Arrow options see
        https://matplotlib.org/stable/_images/quiver_sizes.svg
      - arrow_head_width (float) : Arrow options see
        https://matplotlib.org/stable/_images/quiver_sizes.svg
      - arrow_head_length (float) : Arrow options see
        https://matplotlib.org/stable/_images/quiver_sizes.svg
      - arrow_headaxis_length (float) : Arrow options see
        https://matplotlib.org/stable/_images/quiver_sizes.svg
      - scalarbar : (bool) Print colorbar below figure
      - n_ticks : (int) Number of Tick mars
      - vmin : (float) Minimum Value - Required for colorbar
      - vmax : (float) Maximum Value - Required for colorbar
      - cmap : (color map string) Defaults to "jet"
      - scalarbar_width : (float) Width of the colorbar relative to picture
      - scalarbar_aspect_ratio : (float) Aspect ratio of colorbar
      - scalarbar_offset : (float) reserved space below picture
      - scalarbar_stroke_width : (float) boundary width
      - scalarbar_tick_distance: (float)distance from tick marks to colorbar
      - scalarbar_font_size : (float)

    Returns
    -------
    None
    """
    from splinepy.spline import Spline

    # Check arguments
    if not isinstance(fname, str):
        raise ValueError("fname argument must be string")
    if not isinstance(box_margins, (int, float)):
        raise ValueError("fname argument must be string")

    # Check if user passed list
    if (len(objects_to_plot) == 1) and isinstance(objects_to_plot[0], list):
        objects_to_plot = objects_to_plot[0]

    for object in objects_to_plot:
        if isinstance(object, Spline):
            if not 1 <= object.para_dim <= 2:
                raise ValueError("Only lines and surfaces supported")
            if not object.dim == 2:
                raise ValueError("Only 2D geometries supported")
        elif isinstance(object, _Vertices):
            if not object.vertices.shape[1] == 2:
                raise ValueError("Only 2D geometries supported")
        else:
            raise ValueError("Unexpected spline or gustaf type")

    # Determine bounding box of all elements
    ctps_bounds = _np.nan * _np.ones((2, 2))
    for object in objects_to_plot:
        if isinstance(object, Spline):
            l_bounds = object.control_point_bounds
        else:
            l_bounds = object.bounds()
        ctps_bounds[0, :] = _np.fmin(l_bounds[0, :], ctps_bounds[0, :])
        ctps_bounds[1, :] = _np.fmax(l_bounds[1, :], ctps_bounds[1, :])

    # Check if user requested scalarbar
    scalarbar = kwargs.get("scalarbar", False)

    # These are all scalarbar positioning statements that can be user-modified
    kwargs["scalarbar_offset"] = kwargs.get("scalarbar_offset", 0.1)
    scalarbar_offset = kwargs["scalarbar_offset"]
    # Check if required arguments have been passed
    if scalarbar and (
        (kwargs.get("vmin", None) is None)
        or (kwargs.get("vmax", None) is None)
    ):
        raise ValueError(
            "`vmin` and `vmax` must be passed alon with scalarbar to ensure"
            " same color scheme for all splines"
        )

    # Initialize svg element
    svg_data = _ET.Element("svg")

    # licked it so its ours
    svg_data.insert(
        1,
        _ET.Comment(
            "generated by splinepy https://github.com/tataratat/splinepy"
        ),
    )

    # Set the box margins and svg options globally
    box_size = [
        ctps_bounds[1, 0] - ctps_bounds[0, 0] + 2 * box_margins,
        ctps_bounds[1, 1] - ctps_bounds[0, 1] + 2 * box_margins,
    ]
    # Make space for scalarbar
    if scalarbar:
        box_size[1] *= 1 + scalarbar_offset * 2

    # Convert to tuple to prevent overwrite
    box_size = tuple(box_size)

    svg_data.attrib["viewBox"] = f"0 0 {box_size[0]} {box_size[1]}"
    svg_data.attrib["height"] = str(400)
    svg_data.attrib["width"] = str(400 / box_size[1] * box_size[0])
    background = kwargs.get("background", "white")
    if background is not None:
        svg_data.attrib["style"] = f"background-color:{background}"
    svg_data.attrib["xmlns"] = "http://www.w3.org/2000/svg"
    svg_data.attrib["xmlns:xlink"] = "http://www.w3.org/1999/xlink"

    # Determine offsets
    box_min_x = ctps_bounds[0, 0] - box_margins
    box_max_y = ctps_bounds[1, 1] + box_margins

    # Quiver plots should always be grouped in the top layer
    splines_group = _ET.SubElement(svg_data, "g", id="spline_list")
    quiver_list = _ET.SubElement(svg_data, "g", id="quiver_list")

    # Write splines in svg file
    for i, object in enumerate(objects_to_plot):
        # Copy spline options from keyword arguments
        # (taken from visualize to ensure conformity)
        orig_show_options = None
        if kwargs:
            orig_show_options = object.show_options
            object._show_options = object.__show_option__(object)
            orig_show_options.copy_valid_options(object.show_options)
            for key, value in kwargs.items():
                try:
                    object.show_options[key] = value
                except BaseException:
                    continue

        if isinstance(object, Spline):
            # Put every spline and its dedicated information into a new
            # dedicated group
            spline_group = _ET.SubElement(
                splines_group, "g", id="spline" + str(i)
            )
            _export_spline(
                object,
                spline_group,
                box_min_x,
                box_max_y,
                tolerance=tolerance,
                **kwargs,
            )
            _export_control_mesh(
                object, spline_group, box_min_x, box_max_y, **kwargs
            )
            _quiver_plot(
                spline=object,
                svg_spline_element=quiver_list,
                box_min_x=box_min_x,
                box_max_y=box_max_y,
                **kwargs,
            )
        else:
            gus_grounp = _ET.SubElement(svg_data, "g", id="gus_obj_" + str(i))
            _export_gustaf_object(
                object, gus_grounp, box_min_x, box_max_y, **kwargs
            )

        # set original options back
        if orig_show_options is not None:
            object._show_options = orig_show_options
    if scalarbar:
        _add_scalar_bar(svg_data, box_size, **kwargs)

    # Dump into file
    if indent and hasattr(_ET, "indent"):
        # Pretty printing xml with indent only exists in version > 3.9
        _ET.indent(svg_data)

    file_content = _ET.tostring(svg_data)
    with open(fname, "wb") as f:
        f.write(file_content)
    pass



import gustaf as gus
import numpy as np

import splinepy

if __name__ == "__main__":
    # 1. Curve fitting
    # Create a function that creates the points for approximation
    def f(x):
        return x**4 - 2 * x**2 + x + 0.2 * np.sin(10 * x)

    def evaluate_approximation(spline, points, n_initial_guess=500):
        # Calculate proximities between the spline and the given points
        prox_r = spline.proximities(
            queries=points,
            initial_guess_sample_resolutions=[n_initial_guess]
            * spline.para_dim,
            return_verbose=True,
            nthreads=1,
        )
        # Just evaluate the distance
        distances = prox_r[3]
        return np.max(distances), np.mean(distances)

    n_sample = 25
    n_resolution = 3 * n_sample
    x = np.linspace(-1, 1, n_sample)
    fitting_points = np.vstack([x, f(x)]).T.reshape(-1, 2)
    pts = gus.create.vertices.Vertices(fitting_points)
    pts.show_options["c"] = "blue"
    pts.show_options["r"] = 10

    # fit_curve creates Bspline where the n_control_points = n_fitting_points
    # if no further information is given
    interpolated_curve, _ = splinepy.helpme.fit.curve(
        fitting_points=fitting_points, degree=2
    )

    # In curve approximation the number of control points must be chosen
    n_cps = 10
    approximated_curve, _ = splinepy.helpme.fit.curve(
        fitting_points=fitting_points, degree=3, n_control_points=n_cps
    )

    interp_errs = evaluate_approximation(interpolated_curve, fitting_points)
    approx_errs = evaluate_approximation(approximated_curve, fitting_points)

    gus.show(
        [
            f"Interpolated curve \nMax. distance {interp_errs[0]:.3f}, "
            f"mean distance {interp_errs[1]:.3f}",
            interpolated_curve,
            pts,
        ],
        [
            f"Approximated curve with {n_cps} CPs \nMax. distance "
            f"{approx_errs[0]:.3f}, mean distance {approx_errs[1]:.3f}",
            approximated_curve,
            pts,
        ],
    )

    approximated_curves = []

    # The number of control points affects the approximation error
    for n_cps in range(n_sample, 5, -5):
        approximated_curve, _ = splinepy.helpme.fit.curve(
            fitting_points, degree=3, n_control_points=n_cps
        )

        max_error, mean_error = evaluate_approximation(
            approximated_curve, fitting_points
        )
        approximated_curves.append(
            [n_cps, approximated_curve, [max_error, mean_error]]
        )

    gus_list = [
        [
            f"PY: Approximated curve with {cps} CPs \n Max. "
            f"distance {errs[0]:.2f}, mean distance {errs[1]:.2f}",
            curve,
            pts,
        ]
        for (cps, curve, errs) in approximated_curves
    ]
    gus.show(*gus_list)

    # 2. Surface fitting
    n_sample = [20, 20]
    resolution = [3 * n for n in n_sample]

    # Create dataset
    queries = splinepy.cartesian_product(
        [np.linspace(-1, 1, n) for n in n_sample]
    )

    def target_function(x):
        return np.hstack(
            [
                x,
                -0.9 * (x[:, :1] ** 3)
                + 0.7 * x[:, 1:2] ** 2
                + x[:, :1] * x[:, 1:2]
                + 0.1 * np.sin(5 * x[:, :1]),
            ]
        )

    fitting_points = target_function(queries)

    interpolated_surface, _ = splinepy.helpme.fit.surface(
        fitting_points=fitting_points,
        size=n_sample,
        n_control_points=n_sample,
        degrees=[3, 2],
    )

    n_cps = [int(0.75 * n_sample[0]), int(0.6 * n_sample[1])]

    approximated_surface, _ = splinepy.helpme.fit.surface(
        fitting_points=fitting_points,
        size=n_sample,
        n_control_points=n_cps,
        degrees=[3, 2],
        centripetal=True,
    )

    n_cps_int = interpolated_surface.control_mesh_resolutions
    max_error_int, mean_error_int = evaluate_approximation(
        interpolated_surface, fitting_points
    )

    n_cps = approximated_surface.control_mesh_resolutions
    max_error, mean_error = evaluate_approximation(
        approximated_surface, fitting_points
    )

    pts = gus.create.vertices.Vertices(fitting_points)
    pts.show_options["c"] = "blue"
    pts.show_options["r"] = 10

    gus.show(
        [
            f"Interpolated surface with ({n_cps_int}) CPs\n"
            f"Max. Distance {max_error_int:.3f},"
            f"mean distance: {mean_error_int:.3f}",
            interpolated_surface,
            pts,
        ],
        [
            f"Approximated surface with ({n_cps}) CPs\n"
            f"Max. Distance {max_error:.3f},"
            f"mean distance: {mean_error:.3f}",
            approximated_surface,
            pts,
        ],
    )

    approximated_surfaces = []
    # The approximation error changes with the number of control points
    for r in range(0, min(n_sample) - 2, 5):
        approximated_surface, _ = splinepy.helpme.fit.surface(
            fitting_points=fitting_points,
            size=n_sample,
            n_control_points=[n_sample[0] - int(r / 2), n_sample[1] - r],
            degrees=[3, 2],
            knot_vectors=None,
        )

        n_cps = approximated_surface.control_mesh_resolutions

        max_error, mean_error = evaluate_approximation(
            approximated_surface, fitting_points
        )

        approximated_surfaces.append(
            [n_cps, approximated_surface, [max_error, mean_error]]
        )

    gus_list = [
        [
            f"Approximated surface with {cps} CPs\n Max. distance "
            f"{errs[0]:.3f}, mean distance {errs[1]:.3f}",
            surface,
            pts,
        ]
        for (cps, surface, errs) in approximated_surfaces
    ]
    gus.show(*gus_list)

    # Unevenly spaced ctps
    uneven_queries = splinepy.cartesian_product(
        [(np.linspace(0, 1, 15)), np.square(np.linspace(0, 1, 20))]
    )
    fitting_points = target_function(uneven_queries)
    interpolated_surface, _ = splinepy.helpme.fit.surface(
        fitting_points=fitting_points,
        size=[15, 20],
        n_control_points=[4, 4],
        degrees=[3, 3],
        knot_vectors=[[0, 0, 0, 0, 1, 1, 1, 1], [0, 0, 0, 0, 1, 1, 1, 1]],
    )
    pts = gus.create.vertices.Vertices(fitting_points)
    pts.show_options["c"] = "blue"
    pts.show_options["r"] = 10
    splinepy.show(interpolated_surface, pts)



r"""
In this file we will prototype a collocation example (in the simplest way),
that shows how to implement a Poisson problem in the form
.. math::
        -\Delta u = f
with source term f=1

The refinement will only be applied to the solution field, to make the
calculations more efficient
"""

import numpy as np
from scipy.sparse import linalg

import splinepy as sp

# Test Case
n_refine = 15


# Source Function
def source_function(x):
    return np.ones(x.shape[0])


# Define the Geometry
geometry = sp.BSpline(
    degrees=[2, 2],
    control_points=[
        [0.0, 0.0],
        [1.0, 0.5],
        [2.0, 0.2],
        [0.5, 1.5],
        [1.0, 1.5],
        [1.5, 1.5],
        [0.0, 3.0],
        [1.0, 2.5],
        [2.0, 3.0],
    ],
    knot_vectors=[[0, 0, 0, 1, 1, 1], [0, 0, 0, 1, 1, 1]],
)

# Define the solution field
solution_field = sp.BSpline(
    degrees=geometry.degrees,
    control_points=np.ones((geometry.control_points.shape[0], 1)),
    knot_vectors=geometry.knot_vectors,
)


# Use refinement
solution_field.elevate_degrees([0, 1, 0, 1])
new_knots = np.linspace(1 / n_refine, 1, n_refine, endpoint=False)
solution_field.insert_knots(0, new_knots)
solution_field.insert_knots(1, new_knots)


# Get greville points and geometric values
evaluation_points = solution_field.greville_abscissae()
mapper = solution_field.mapper(reference=geometry)
sp_laplacian = sp.utils.data.make_matrix(
    *mapper.basis_laplacian_and_support(evaluation_points),
    solution_field.cps.shape[0],
    as_array=False,
)


# Evaluate RHS function
physical_points = geometry.evaluate(evaluation_points)
rhs = source_function(physical_points)

# Set dirichlet values (identify boundary nodes and set matrix to identity rhs
# to 0)
boundary_dof_ids = np.concatenate(
    (
        solution_field.multi_index[0, :],
        solution_field.multi_index[-1, :],
        solution_field.multi_index[1:-1, 0],
        solution_field.multi_index[1:-1, -1],
    )
)
sp_laplacian[boundary_dof_ids] *= 0  # this does not change sparsity pattern
sp_laplacian[boundary_dof_ids, boundary_dof_ids] = 1

rhs[boundary_dof_ids] = 0.0

# Solve linear system
solution_field.control_points[:] = linalg.spsolve(-sp_laplacian, rhs).reshape(
    -1, 1
)


# Plot geometry and field
geometry.spline_data["field"] = solution_field
geometry.show_options["data"] = "field"
geometry.show_options["cmap"] = "jet"
geometry.show_options["lighting"] = "off"
geometry.show_options["scalarbar"] = True
geometry.show(knots=True, control_points=False)



import numpy as np

# for parametrization used fixtures
all_2p2d_splines = (
    "bezier_2p2d",
    "rational_bezier_2p2d",
    "bspline_2p2d",
    "nurbs_2p2d",
)


def convert_and_compare_samples(spline, conversion_list):
    for convert_to in conversion_list:
        converted = getattr(spline, convert_to)

        sample_res = [3] * spline.para_dim
        ref_sample = spline.sample(sample_res)

        # check same sample result
        assert np.allclose(ref_sample, converted.sample(sample_res))

        # check same type
        assert type(converted).__qualname__.lower().startswith(convert_to)


def test_bezier_conversion(bezier_2p2d):
    test_conversions = ["bezier", "rationalbezier", "bspline", "nurbs"]
    convert_and_compare_samples(bezier_2p2d, test_conversions)


def test_rational_bezier_conversion(rational_bezier_2p2d):
    test_conversions = ["rationalbezier", "nurbs"]
    convert_and_compare_samples(rational_bezier_2p2d, test_conversions)


def test_bspline_conversion(bspline_2p2d):
    test_conversions = ["bspline", "nurbs"]
    convert_and_compare_samples(bspline_2p2d, test_conversions)


def test_nurbs_conversion(nurbs_2p2d):
    test_conversions = ["nurbs"]
    convert_and_compare_samples(nurbs_2p2d, test_conversions)



"""Call all *.py files in the working directory.
"""

import glob
import itertools
import pathlib
import subprocess
import sys

if __name__ == "__main__":
    files_not_completed = []
    examples_root_dir = pathlib.Path(__file__).parent.absolute()
    for file in itertools.chain(
        glob.glob(str(examples_root_dir / "*.py")),
        glob.glob(str(examples_root_dir / "iga/*.py")),
    ):
        if "run_all_examples.py" in file:
            continue
        print(f"Calling {file}")
        proc_return = subprocess.run([sys.executable, file], check=False)
        if proc_return.returncode != 0:
            files_not_completed.append(file)
    if len(files_not_completed) > 0:
        print(
            f"Failed to call {len(files_not_completed)} files: "
            f"{files_not_completed}."
        )
        sys.exit(1)

    print("All files completed successfully.")
    sys.exit(0)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
import napf

# sys.path.insert(0, os.path.abspath("../../"))


# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "napf"
copyright = "2022, Jaewook Lee"
author = "Jaewook Lee"
release = napf.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

source_suffix = {
    ".rst": "restructuredtext",
    ".txt": "markdown",
    ".md": "markdown",
}

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_mdinclude",
]


templates_path = ["_templates"]
exclude_patterns = []

pygments_style = "sphinx"


# -- Options for HTML output -------------------------------------------------

html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]


autodoc_default_options = {
    "autosummary": True,
}

autosummary_context = {
    "skipmethods": ["__init__"],
}



import numpy as np

import splinepy


def test_fit_curve_2d_interpolation():
    n_sample = 20
    resolution = [3 * n_sample]
    degree = 2

    x = np.linspace(-2, 2, n_sample)

    def f(x):
        return x**2

    fitting_points = np.vstack((x, f(x))).T.reshape(-1, 2)

    interpolation_spline, _ = splinepy.helpme.fit.curve(
        fitting_points=fitting_points, degree=degree
    )
    interpolated_points = interpolation_spline.sample(resolutions=resolution)

    assert np.allclose(
        f(interpolated_points[:, 0]),
        interpolated_points[:, 1],
        atol=1e-3,
    )


def test_fit_curve_2d_approximation():
    n_sample = 20
    resolution = [3 * n_sample]
    degree = 2

    x = np.linspace(-2, 2, n_sample)

    def f(x):
        return x**2

    fitting_points = np.vstack((x, f(x))).T.reshape(-1, 2)

    approximated_spline, _ = splinepy.helpme.fit.curve(
        fitting_points=fitting_points,
        degree=degree,
        n_control_points=n_sample - 5,
    )
    approximated_points = approximated_spline.sample(resolution)

    assert np.allclose(
        approximated_points[:, 1],
        f(approximated_points[:, 0]),
        atol=1e-2,
    )


def test_fit_surface_3d_interpolation():
    sample_size = [15, 10]
    x = [np.linspace(-2, 2, n) for n in sample_size]
    xx, yy = np.meshgrid(x[0], x[1])

    def f(x, y):
        return x**2 + y**2

    fitting_points = np.vstack(
        (xx.flatten(), yy.flatten(), f(xx, yy).flatten())
    ).T

    bspline, _ = splinepy.helpme.fit.surface(
        fitting_points=fitting_points,
        size=[sample_size[0], sample_size[1]],
        n_control_points=[sample_size[0], sample_size[1]],
        degrees=[2, 2],
    )
    approximated_points = bspline.sample(sample_size)

    assert np.allclose(
        f(approximated_points[:, 0], approximated_points[:, 1]),
        approximated_points[:, -1],
        atol=1e-2,
    )


def test_fit_surface_3d_approximation():
    sample_size = [10, 15]
    x = [np.linspace(-1, 1, n) for n in sample_size]
    xx, yy = np.meshgrid(x[0], x[1])

    def f(x, y):
        return 0.5 * (x**2 + y**2)

    fitting_points = np.vstack(
        (xx.flatten(), yy.flatten(), f(xx, yy).flatten())
    ).T

    bspline, _ = splinepy.helpme.fit.surface(
        fitting_points=fitting_points,
        size=[sample_size[0], sample_size[1]],
        n_control_points=[sample_size[0] - 2, sample_size[1] - 2],
        degrees=[3, 3],
    )
    approximated_points = bspline.sample(sample_size)

    assert np.allclose(
        f(approximated_points[:, 0], approximated_points[:, 1]),
        approximated_points[:, -1],
        atol=1e-3,
    )



import os
import tempfile

import splinepy


def test_mfem_export(to_tmpf, are_stripped_lines_same):
    """
    Test MFEM export routine
    """
    # Define some splines
    bez_el0 = splinepy.Bezier(
        degrees=[1, 1], control_points=[[0, 0], [1, 0], [0, 1], [1, 1]]
    )
    rbz_el1 = splinepy.RationalBezier(
        degrees=[1, 1],
        control_points=[[1, 0], [2, 0], [1, 1], [2, 1]],
        weights=[1, 1, 1, 1],
    )
    bsp_el2 = splinepy.BSpline(
        degrees=[1, 1],
        control_points=[[0, 1], [1, 1], [0, 2], [1, 2]],
        knot_vectors=[[0, 0, 1, 1], [0, 0, 1, 1]],
    )
    nur_el3 = splinepy.NURBS(
        degrees=[1, 1],
        control_points=[[1, 1], [2, 1], [1, 2], [2, 2]],
        weights=[1, 1, 1, 1],
        knot_vectors=[[0, 0, 1, 1], [0, 0, 1, 1]],
    )

    # Make it more tricky
    bez_el0.elevate_degrees(0)
    bsp_el2.elevate_degrees(0)

    bsp_el2.insert_knots(1, [0.5])
    nur_el3.insert_knots(1, [0.5])

    # Test Output
    with tempfile.TemporaryDirectory() as tmpd:
        tmpf = to_tmpf(tmpd)
        splinepy.io.mfem.export_cartesian(
            tmpf, [bez_el0, bsp_el2, nur_el3, rbz_el1]
        )

        with open(tmpf) as tmp_read, open(
            os.path.dirname(__file__) + "/../data/mfem_cartesian_2d.mesh"
        ) as base_file:
            assert are_stripped_lines_same(
                base_file.readlines(), tmp_read.readlines(), True
            )

    # Test Also 3D Meshes
    bez_el0 = splinepy.Bezier(
        degrees=[1, 1, 1],
        control_points=[
            [0, 0, 0],
            [1, 0, 0],
            [0, 1, 0],
            [1, 1, 0],
            [0, 0, 1],
            [1, 0, 1],
            [0, 1, 1],
            [1, 1, 1],
        ],
    )
    rbz_el1 = splinepy.RationalBezier(
        degrees=[1, 1, 1],
        control_points=[
            [1, 0, 0],
            [2, 0, 0],
            [1, 1, 0],
            [2, 1, 0],
            [1, 0, 1],
            [2, 0, 1],
            [1, 1, 1],
            [2, 1, 1],
        ],
        weights=[1] * 8,
    )
    bsp_el2 = splinepy.BSpline(
        degrees=[1, 1, 1],
        control_points=[
            [0, 1, 0],
            [1, 1, 0],
            [0, 2, 0],
            [1, 2, 0],
            [0, 1, 1],
            [1, 1, 1],
            [0, 2, 1],
            [1, 2, 1],
        ],
        knot_vectors=[[0, 0, 1, 1]] * 3,
    )
    nur_el3 = splinepy.NURBS(
        degrees=[1, 1, 1],
        control_points=[
            [1, 1, 0],
            [2, 1, 0],
            [1, 2, 0],
            [2, 2, 0],
            [1, 1, 1],
            [2, 1, 1],
            [1, 2, 1],
            [2, 2, 1],
        ],
        weights=[1] * 8,
        knot_vectors=[[0, 0, 1, 1]] * 3,
    )

    # Make it more tricky
    bez_el0.elevate_degrees(0)
    bsp_el2.elevate_degrees(0)
    bsp_el2.insert_knots(1, [0.5])
    nur_el3.insert_knots(1, [0.5])
    for s in [bez_el0, bsp_el2, nur_el3, rbz_el1]:
        s.elevate_degrees(2)

    # Test output
    with tempfile.TemporaryDirectory() as tmpd:
        tmpf = to_tmpf(tmpd)
        splinepy.io.mfem.export_cartesian(
            tmpf, [bez_el0, bsp_el2, nur_el3, rbz_el1]
        )

        with open(tmpf) as tmp_read, open(
            os.path.dirname(__file__) + "/../data/mfem_cartesian_3d.mesh"
        ) as base_file:
            assert are_stripped_lines_same(
                base_file.readlines(), tmp_read.readlines(), True
            )



import numpy as np
import pytest

# fixtures used
all_2p2d_splines = (
    "rational_bezier_2p2d",
    "bezier_2p2d",
    "bspline_2p2d",
    "nurbs_2p2d",
)


@pytest.mark.parametrize("splinetype", all_2p2d_splines)
def test_queries_inside_spline_initial_guess_with_kdt(
    request, splinetype, np_rng
):
    """
    Initial guess made with kdt. Mid-point as initial guess tends to fail,
    so excluded from test.
    """
    spline = request.getfixturevalue(splinetype)
    # form parametric queries
    # -> parametric space is [0, 1], so no further manipulation
    para_q = np_rng.random((10, spline.para_dim))

    # form physical queries
    phys_q = spline.evaluate(para_q)

    # proximity - single thread exe
    prox_r = spline.proximities(
        queries=phys_q,
        initial_guess_sample_resolutions=[10] * spline.para_dim,
        nthreads=1,
        return_verbose=True,
    )

    assert np.allclose(
        para_q, prox_r[0]
    ), f"WRONG proximity query for {spline.whatami}"



from functools import wraps as _wraps

import numpy as _np

from splinepy.utils.data import cartesian_product as _cartesian_product


def _get_integral_measure(spline):
    """
    Determines the appropriate measure to be used in integration

    If the spline dimension matches its parametric dimension it will return a
    Callable in the form

    .. math::
        \\mathcal{J}_S = det(\\mathbf(J))

    If the physical dimension is bigger then the paramtric dimension it will
    return

    .. math::
        \\mathcal{J}_S = \big( det(\\mathbf(J)^T \\mathbf(J)\big)^{0.5}

    Parameters
    ----------
    spline : Spline / Multipatch
      For parametric and physical dimension

    Returns
    -------
    measure : Callable
      single patch only
    """
    # Check dimensionality
    if spline.dim == spline.para_dim:

        def measure(spline_patch, positions):
            return _np.linalg.det(spline_patch.jacobian(positions))

        return measure

    elif spline.dim > spline.para_dim:

        def measure(spline_patch, positions):
            jacs = spline_patch.jacobian(positions)
            return _np.sqrt(
                _np.linalg.det(_np.einsum("oji,ojk->oik", jacs, jacs))
            )

        return measure

    else:
        raise ValueError("`Volume` not supported if para_dim > dim")


def _get_quadrature_information(spline, orders=None):
    """
    Select appropriate integration order (gauss-legendre)

    Determinante of a polynomial spline with para_dim==dim has degree

    .. math::
        p_i^{det} = n_{dim} \\cdot p_i - 1

    cf. [Mantzaflaris et al., 2017,
    DOI:http://dx.doi.org/10.1016/j.cma.2016.11.013]

    The same order approximation will also be used for the metric

    .. math::
        \\mathcal{J}_S = \big( det(\\mathbf(J)^T \\mathbf(J)\big)^{0.5}

    Parameters
    ----------
    spline : Spline
      Spline for integration
    orders : array-like (optional)
      Orders along every parametric dimension

    Returns
    -------
    positions : np.ndarray
      quadrature position in unit-square
    weights : np.ndarray
      quadrature weights
    """

    # Determine integration points
    positions = []
    weights = []

    # Determine integration orders
    if orders is None:
        expected_degree = spline.degrees * spline.para_dim + 1
        if spline.is_rational:
            expected_degree += 2
            spline._logd(
                "Integration on rational spline is only approximation"
            )

        # Gauss-Legendre is exact for polynomials 2*n-1
        quad_orders = _np.ceil((expected_degree + 1) * 0.5).astype("int")
    else:
        quad_orders = _np.ascontiguousarray(orders, dtype=int).flatten()
        if quad_orders.size != spline.para_dim:
            raise ValueError(
                "Integration order must be array of size para_dim"
            )

    for order in quad_orders:
        # Get legendre quadratuer points
        pos, ws = _np.polynomial.legendre.leggauss(order)
        # from (-1,1) to (0,1)
        positions.append(pos * 0.5 + 0.5)
        weights.append(ws * 0.5)

    # summarize integration points
    return (
        _cartesian_product(positions),
        _cartesian_product(weights).prod(axis=1),
    )


def volume(spline, orders=None):
    r"""Compute volume of a given spline

    Parameters
    ----------
    spline : Spline
        (self if called via integrator)
      splinepy - spline type
    orders : array-like (optional)
      order for gauss quadrature

    Returns
    -------
    volume : float
      Integral of dim-dimensional object
    """
    from splinepy.spline import Spline as _Spline

    # Check i_nput type
    if not isinstance(spline, _Spline):
        raise NotImplementedError("integration only works for splines")

    # Retrieve aux info
    meas = _get_integral_measure(spline)
    positions, weights = _get_quadrature_information(spline, orders)

    # Calculate Volume
    if spline.has_knot_vectors:
        volume = _np.sum(
            [
                _np.sum(meas(b, positions) * weights)
                for b in spline.extract.beziers()
            ]
        )
    else:
        volume = _np.sum(meas(spline, positions) * weights)
    return volume


def parametric_function(
    spline,
    function,
    orders=None,
):
    """Integrate a function defined within the parametric domain

    Parameters
    ----------
    spline : Spline
        (self if called via integrator)
    function : Callable
    orders : optional

    Returns
    -------
    integral : np.ndarray
    """
    from splinepy.spline import Spline as _Spline

    # Check i_nput type
    if not isinstance(spline, _Spline):
        raise NotImplementedError("integration only works for splines")

    # Retrieve aux info
    meas = _get_integral_measure(spline)
    positions, weights = _get_quadrature_information(spline, orders)

    # Calculate Volume
    if spline.has_knot_vectors:
        # positions must be mapped into each knot-element
        para_view = spline.create.parametric_view(axes=False)

        # get initial shape
        initial = function([positions[0]])
        result = _np.zeros(initial.shape[1])
        for bezier_element in para_view.extract.beziers():
            quad_positions = bezier_element.evaluate(positions)
            result += _np.einsum(
                "i...,i,i->...",
                function(quad_positions),
                meas(spline, quad_positions),
                weights,
                optimize=True,
            )

    else:
        result = _np.einsum(
            "i...,i,i->...",
            function(positions),
            meas(spline, positions),
            weights,
            optimize=True,
        )
    return result


def physical_function(
    function,  # noqa ARG001
    orders,  # noqa ARG001
):
    """Integrate a function defined within the physical domain"""
    raise NotImplementedError(
        "Function not implemented yet. Please feel free to write an issue, if "
        "you need it: github.com/tatarata/splinepy/issues"
    )


class Integrator:
    """Helper class to integrate some values on a given spline geometry

    Examples
    --------

    .. code-block:: python

        splinepy.helpme.integrate.volume()
        # Equivalent to
        spline.integrate.volume()

    Parameters
    ----------
    spline : Spline
      Spline parent
    """

    __slots__ = ("_helpee",)

    def __init__(self, spl):
        self._helpee = spl

    @_wraps(volume)
    def volume(self, *args, **kwargs):
        return volume(self._helpee, *args, **kwargs)

    @_wraps(parametric_function)
    def parametric_function(self, *args, **kwargs):
        return parametric_function(self._helpee, *args, **kwargs)



import numpy as np
import pytest

# frequently used fixtures
all_splinetypes = ("bspline_3p3d", "nurbs_3p3d", "bspline_2p2d", "nurbs_2p2d")


@pytest.mark.parametrize("splinetype", all_splinetypes)
def test_knot_vectors(splinetype, request):
    spline = request.getfixturevalue(splinetype)
    # test unique_knots
    copy_knot_vectors = spline.knot_vectors[:]
    unique_knots = [np.unique(ckvs) for ckvs in copy_knot_vectors]

    for uk, uk_fct in zip(unique_knots, spline.unique_knots):
        assert np.allclose(uk, uk_fct)

    # test knot_multiplicities
    multiplicity = [
        np.unique(ckvs, return_counts=True) for ckvs in copy_knot_vectors
    ]

    for m_u, m_fct in zip(multiplicity, spline.knot_multiplicities):
        assert np.allclose(m_u[1], m_fct)

    # test knot_vector creation
    for u_kv, kn_m, spl_kv in zip(
        spline.unique_knots,
        spline.knot_multiplicities,
        spline.knot_vectors,
    ):
        assert np.allclose(np.array(spl_kv), np.repeat(u_kv, kn_m))



"""
iges io.
"""

from splinepy import splinepy_core as _splinepy_core


def load(fname):
    """
    Read spline in `.iges` form.

    Parameters
    -----------
    fname: str

    Returns
    --------
    splines: list
      Spline Type defined in NAME_TO_TYPE
    """
    return _splinepy_core.read_iges(fname)


def export(fname, splines):
    """
    Save splines as `.iges`.

    Parameters
    -----------
    fname: str
    spline: list
      list of Splines

    Returns
    --------
    None
    """
    return _splinepy_core.export_iges(fname, splines)



"""
xml io.
Keyword follows RWTH CATS's spline format.
Possibly will be turned into pure python io.
"""

import xml.etree.ElementTree as _ET
from sys import version as _python_version

import numpy as _np

from splinepy.io import ioutils as _ioutils
from splinepy.utils.log import debug as _debug
from splinepy.utils.log import warning as _warning

# List of spline keywords (bundled here in case they change - many unused)
CATS_XML_KEY_WORDS = {
    "spline_type": "SplineType",  # useless keyword (always one)
    "patch": "SplineEntry",  # key for new patch in List
    "n_patches": "NumberOfSplines",  # number of patches
    "para_dim": "splDim",  # Parametric dimension
    "dim": "spaceDim",  # Spatial dimension
    "field_dim": "numOfCntrlPntVars",  # Number of control point variables
    "n_ctps": "numCntrlPnts",  # Number of control points
    "n_element_vars": "numOfEleVars",  # Number of element variables
    "periodic": "closed",  # Is the spline closed? No[0] or yes[1] for each
    # parameter direction
    "coeff_names": "cntrlPntVarNames",  # Names of coefficients
    "control_points": "cntrlPntVars",  # Actual control points
    "conforming_ctps": "connectedCntrlPnts",  # Interfaces and conforming
    # control points
    "weights": "wght",  # Weights (every spline is a nurbs)
    "degrees": "deg",  # Degrees vector
    "knot_vectors": "kntVecs",  # Vector of knot vectors
    "knot_vector": "kntVec",  # individual knot vector
    "spline_list": "SplineList",  # Global list of patches
    "element_variables": "elmentVars",  # Variables defined on the element
}


def load(fname):
    """
    Read spline in `.xml` form.
    RWTH CATS spline format.

    Parameters
    -----------
    fname: str

    Returns
    --------
    splines: Multipatch object
      Spline Type defined in NAME_TO_TYPE
    """

    def _read_spline(xml_element):
        spline_dict = {}
        if not xml_element.tag.startswith(CATS_XML_KEY_WORDS["patch"]):
            _debug(
                f"Found unexpected keyword {xml_element.tag}, which will be "
                f"ignored"
            )
        # Read required information from Header in xml element

        dim = int(xml_element.attrib.get(CATS_XML_KEY_WORDS["dim"], -1))
        para_dim = int(
            xml_element.attrib.get(CATS_XML_KEY_WORDS["para_dim"], -1)
        )
        ctps_dim = int(
            xml_element.attrib.get(CATS_XML_KEY_WORDS["field_dim"], -1)
        )
        n_ctps = int(xml_element.attrib.get(CATS_XML_KEY_WORDS["n_ctps"], -1))
        periodic_keys = xml_element.attrib.get(
            CATS_XML_KEY_WORDS["periodic"], ""
        ).split()
        any_is_periodic = any(int(key) == 1 for key in periodic_keys)
        if -1 in [dim, para_dim, ctps_dim, n_ctps]:
            raise ValueError(
                f"Not enough information provided for xml-element "
                f"{xml_element}"
            )
        if ctps_dim > dim:
            _debug(
                "Spline seems to contain field information, but field "
                "retrieval is not implemented. All non-geometry info will be "
                "ignored"
            )
        if any_is_periodic:
            raise ValueError(
                "Periodic splines are currently unsupported. Detected "
                "periodic dimensions: "
                "{[i for i, key in enumerate(periodic_keys) if int(key) == 1]}"
            )
        for info in xml_element:
            # We ignore most keywords at the moment
            if CATS_XML_KEY_WORDS["coeff_names"] in info.tag:
                _debug(
                    f"Will use the following control point vars as control "
                    f"point coordinates : {info.text.split()[:dim]}"
                )
            # Control points
            elif CATS_XML_KEY_WORDS["control_points"] in info.tag:
                spline_dict["control_points"] = _np.fromstring(
                    info.text.replace("\n", " "), sep=" "
                ).reshape(-1, ctps_dim)[:, :dim]

            # degrees
            elif CATS_XML_KEY_WORDS["degrees"] in info.tag:
                spline_dict["degrees"] = _np.fromstring(
                    info.text.replace("\n", " "), sep=" ", dtype=_np.int64
                )
                if spline_dict["degrees"].size != para_dim:
                    raise ValueError(
                        f"Conflicting information provided. Got "
                        f"{spline_dict['degrees'].size} degrees, but "
                        f"parametric dimension is {para_dim}"
                    )

            # weights
            elif CATS_XML_KEY_WORDS["weights"] in info.tag:
                spline_dict["weights"] = _np.fromstring(
                    info.text.replace("\n", " "), sep=" "
                )

            # knot vectors
            elif CATS_XML_KEY_WORDS["knot_vectors"] in info.tag:
                spline_dict["knot_vectors"] = []
                for child_info in info:
                    if CATS_XML_KEY_WORDS["knot_vector"] not in child_info.tag:
                        _debug("Redundant item in knot_vectors block of xml")
                    spline_dict["knot_vectors"].append(
                        _np.fromstring(
                            child_info.text.replace("\n", " "), sep=" "
                        )
                    )

            # All other keywords will be ignored for the moment
            else:
                _debug(f"Ignoring keyword {info.tag}")

        return spline_dict

    # Parse  XML file
    _debug(f"Parsing xml-file '{fname}' ...")
    root = _ET.parse(fname).getroot()
    _debug("XML-file parsed start conversion")
    list_of_splines = []
    if root.tag.startswith(CATS_XML_KEY_WORDS["spline_list"]):
        if root.attrib.get(CATS_XML_KEY_WORDS["spline_type"], "1") != "1":
            raise ValueError(
                f"Unknown SplineType "
                f"{root.attrib[CATS_XML_KEY_WORDS['spline_type']]}"
            )
        n_patches = int(root.attrib.get(CATS_XML_KEY_WORDS["n_patches"], 0))
        for patch_element in root:
            list_of_splines.append(_read_spline(patch_element))
    else:
        _debug(f"Unused xml-keyword {root.tag}")
        return []

    _debug(f"Found a total of {len(list_of_splines)} " f"BSplines and NURBS")
    if len(list_of_splines) != n_patches:
        raise ValueError(
            f"Found {len(list_of_splines)} splines, but expected to find "
            f"{n_patches} patches"
        )

    # Return splines
    return _ioutils.dict_to_spline(list_of_splines)


def export(fname, spline_list, indent=True, make_rational=True):
    """
    Save spline as `.xml`.

    Parameters
    -----------
    fname: str
      Export file name
    multipatch: Multipatch / list
      Splines to be exported
    indent: bool
      Option for pretty printing
    make_rational : bool
      (default True) export everything as nurbs (discard weights)

    Returns
    --------
    None
    """
    from splinepy.multipatch import Multipatch as _Multipatch
    from splinepy.spline import Spline as _Spline

    # First transform spline-data into a multipatch-data if required
    if issubclass(type(spline_list), _Spline):
        # Transform to list
        spline_list = [spline_list]

    if isinstance(spline_list, _Multipatch):
        # @todo: current implementation only exports geometry, export fields
        if spline_list.fields:
            _warning("Fields are not supported yet")
        # Transform to multipatch
        spline_list = spline_list.patches

    if not isinstance(spline_list, list):
        raise ValueError(
            "export Function expects list for multipatch argument"
        )

    # Start a list of a list of patches
    spline_list_element = _ET.Element(
        CATS_XML_KEY_WORDS["spline_list"],
        SplineType=str(1),
        **{CATS_XML_KEY_WORDS["n_patches"]: str(len(spline_list))},
    )

    # licked it so its ours
    spline_list_element.insert(
        1,
        _ET.Comment(
            "generated by splinepy https://github.com/tataratat/splinepy"
        ),
    )

    new_line_char = "\n" if indent else " "

    # All Splines (patches) are written into the spline list as entries
    for spline in spline_list:
        # Convert to non-bezier type (might make unnecessary copy)
        patch = (
            spline.nurbs
            if spline.is_rational or make_rational
            else spline.bspline
        )

        # Write spline header
        patch_element = _ET.SubElement(
            spline_list_element,
            CATS_XML_KEY_WORDS["patch"],
            **{
                CATS_XML_KEY_WORDS["para_dim"]: str(patch.para_dim),
                CATS_XML_KEY_WORDS["dim"]: str(patch.dim),
                CATS_XML_KEY_WORDS["field_dim"]: str(patch.dim),
                CATS_XML_KEY_WORDS["n_ctps"]: str(patch.cps.shape[0]),
                CATS_XML_KEY_WORDS["n_element_vars"]: str(0),
                CATS_XML_KEY_WORDS["periodic"]: " ".join(
                    [str(0) for _ in range(patch.para_dim)]
                ),
            },
        )

        # Variable names (@todo fields)
        control_point_var_names = _ET.SubElement(
            patch_element,
            CATS_XML_KEY_WORDS["coeff_names"],
        )

        # These are usually hard coded variable names in xns or feafa
        # x y z t seems to only matter. see issue #257
        cp_varnames = [
            "x",
            "y",
            "z",
            "t",
            *[f"x{i}" for i in range(4, patch.dim)],
        ]
        control_point_var_names.text = " ".join(cp_varnames[: patch.dim])

        # Control point variables
        control_points_elements = _ET.SubElement(
            patch_element,
            CATS_XML_KEY_WORDS["control_points"],
        )
        control_points_elements.text = new_line_char.join(
            " ".join(str(xx) for xx in x) for x in patch.cps
        )

        # degrees
        degrees_elements = _ET.SubElement(
            patch_element,
            CATS_XML_KEY_WORDS["degrees"],
        )
        degrees_elements.text = new_line_char.join(
            str(deg) for deg in patch.degrees
        )

        # knot-vectors
        knot_vectors_elements = _ET.SubElement(
            patch_element,
            CATS_XML_KEY_WORDS["knot_vectors"],
        )
        for kv in patch.knot_vectors:
            knot_vector_element = _ET.SubElement(
                knot_vectors_elements,
                CATS_XML_KEY_WORDS["knot_vector"],
            )
            knot_vector_element.text = new_line_char.join(str(k) for k in kv)

        # weights if rational
        if patch.is_rational:
            weights_elements = _ET.SubElement(
                patch_element,
                CATS_XML_KEY_WORDS["weights"],
            )
            weights_elements.text = new_line_char.join(
                str(w) for w in patch.weights.ravel()
            )

    # Beautify and export
    if int(_python_version.split(".")[1]) >= 9 and indent:
        # Pretty printing xml with indent only exists in version > 3.9
        _ET.indent(spline_list_element)

    elif int(_python_version.split(".")[1]) < 9 and indent:
        _debug(
            "Indented xml output is only supported from > python3.9.",
            "Output will not be indented.",
            f"Current python version: {_python_version}",
        )

    file_content = _ET.tostring(spline_list_element)
    with open(fname, "wb") as f:
        f.write(file_content)



import numpy as _np

from splinepy.bezier import Bezier as _Bezier
from splinepy.microstructure.tiles.tile_base import TileBase as _TileBase


class HollowCube(_TileBase):
    """A 3D box with a hollow cube inside.

    .. raw:: html

        <p><a href="../_static/HollowCube.html">Fullscreen</a>.</p>
        <embed type="text/html" width="100%" height="400" src="../_static/HollowCube.html" />

    """  # noqa: E501

    _dim = 3
    _para_dim = 3
    _evaluation_points = _np.array(
        [
            [0.0, 0.5, 0.5],
            [1.0, 0.5, 0.5],
            [0.5, 0.0, 0.5],
            [0.5, 1.0, 0.5],
            [0.5, 0.5, 0.0],
            [0.5, 0.5, 1.0],
            [0.5, 0.5, 0.5],
        ]
    )
    _n_info_per_eval_point = 1

    def create_tile(
        self,
        parameters=None,
        parameter_sensitivities=None,
        **kwargs,  # noqa ARG002
    ):
        """Create a microtile based on the parameters that describe the strut
        thicknesses.


        Parameters
        ----------
        parameters :
          specifies the hole-thickness on the tile interfaces. Last parameter
          is used for center dimensions
        parameter_sensitivities: np.ndarray
          Describes the parameter sensitivities with respect to some design
          variable. In case the design variables directly apply to the
          parameter itself, they evaluate to delta_ij

        Returns
        -------
        microtile_list : list(splines)
        """

        if parameters is None:
            self._logd("Setting parameters to default value (0.2)")
            parameters = (
                _np.ones(
                    (len(self._evaluation_points), self._n_info_per_eval_point)
                ).reshape(-1, 1)
                * 0.2
            )

        self.check_params(parameters)

        if not (
            _np.all(parameters[:, :2] > 0.0)
            and _np.all(parameters[:, :2] < 0.5)
        ):
            raise ValueError("The wall thickness must be in (0.0 and 0.5)")

        if not _np.all(
            (parameters[:, 2:] > _np.deg2rad(-30))
            and (parameters[:, 2:] < _np.deg2rad(30))
        ):
            raise ValueError("Rotation is only allowed between +-30 deg")

        if self.check_param_derivatives(parameter_sensitivities):
            n_derivatives = parameter_sensitivities.shape[2]
            derivatives = []
        else:
            n_derivatives = 0
            derivatives = None

        splines = []

        for i_derivative in range(n_derivatives + 1):
            spline_list = []
            # Constant auxiliary values
            if i_derivative == 0:
                v_zero = 0.0
                v_one = 1.0
                [
                    x_min,
                    x_max,
                    y_min,
                    y_max,
                    z_min,
                    z_max,
                    center,
                ] = parameters[:, 0].flatten()
            else:
                v_zero = 0.0
                v_one = 0.0
                [
                    x_min,
                    x_max,
                    y_min,
                    y_max,
                    z_min,
                    z_max,
                    center,
                ] = parameter_sensitivities[:, 0, i_derivative - 1]

            # x_max_z_max
            spline_list.append(
                _Bezier(
                    degrees=[1, 1, 1],
                    control_points=_np.array(
                        [
                            [v_one - z_max, z_max, v_one],
                            [v_one, v_zero, v_one],
                            [v_one - z_max, v_one - z_max, v_one],
                            [v_one, v_one, v_one],
                            [v_one - center, center, v_one - center],
                            [v_one, x_max, v_one - x_max],
                            [v_one - center, v_one - center, v_one - center],
                            [v_one, v_one - x_max, v_one - x_max],
                        ]
                    ),
                )
            )

            # x_min_z_min
            spline_list.append(
                _Bezier(
                    degrees=[1, 1, 1],
                    control_points=_np.array(
                        [
                            [v_zero, x_min, x_min],
                            [center, center, center],
                            [v_zero, v_one - x_min, x_min],
                            [center, v_one - center, center],
                            [v_zero, v_zero, v_zero],
                            [z_min, z_min, v_zero],
                            [v_zero, v_one, v_zero],
                            [z_min, v_one - z_min, v_zero],
                        ]
                    ),
                )
            )

            # x_min_z_max
            spline_list.append(
                _Bezier(
                    degrees=[1, 1, 1],
                    control_points=_np.array(
                        [
                            [v_zero, v_zero, v_one],
                            [z_max, z_max, v_one],
                            [v_zero, v_one, v_one],
                            [z_max, v_one - z_max, v_one],
                            [v_zero, x_min, v_one - x_min],
                            [center, center, v_one - center],
                            [v_zero, v_one - x_min, v_one - x_min],
                            [center, v_one - center, v_one - center],
                        ]
                    ),
                )
            )

            # x_max_z_min
            spline_list.append(
                _Bezier(
                    degrees=[1, 1, 1],
                    control_points=_np.array(
                        [
                            [v_one - center, center, center],
                            [v_one, x_max, x_max],
                            [v_one - center, v_one - center, center],
                            [v_one, v_one - x_max, x_max],
                            [v_one - z_min, z_min, v_zero],
                            [v_one, v_zero, v_zero],
                            [v_one - z_min, v_one - z_min, v_zero],
                            [v_one, v_one, v_zero],
                        ]
                    ),
                )
            )

            # x_max_y_max
            spline_list.append(
                _Bezier(
                    degrees=[1, 1, 1],
                    control_points=_np.array(
                        [
                            [v_one - y_max, v_one, y_max],
                            [v_one, v_one, v_zero],
                            [v_one - y_max, v_one, v_one - y_max],
                            [v_one, v_one, v_one],
                            [v_one - center, v_one - center, center],
                            [v_one, v_one - x_max, x_max],
                            [v_one - center, v_one - center, v_one - center],
                            [v_one, v_one - x_max, v_one - x_max],
                        ]
                    ),
                )
            )

            # y_max_z_min
            spline_list.append(
                _Bezier(
                    degrees=[1, 1, 1],
                    control_points=_np.array(
                        [
                            [v_zero, v_one, v_zero],
                            [v_one, v_one, v_zero],
                            [y_max, v_one, y_max],
                            [v_one - y_max, v_one, y_max],
                            [z_min, v_one - z_min, v_zero],
                            [v_one - z_min, v_one - z_min, v_zero],
                            [center, v_one - center, center],
                            [v_one - center, v_one - center, center],
                        ]
                    ),
                )
            )

            # x_min_y_max
            spline_list.append(
                _Bezier(
                    degrees=[1, 1, 1],
                    control_points=_np.array(
                        [
                            [v_zero, v_one, v_zero],
                            [y_max, v_one, y_max],
                            [v_zero, v_one, v_one],
                            [y_max, v_one, v_one - y_max],
                            [v_zero, v_one - x_min, x_min],
                            [center, v_one - center, center],
                            [v_zero, v_one - x_min, v_one - x_min],
                            [center, v_one - center, v_one - center],
                        ]
                    ),
                )
            )

            # y_max_z_max
            spline_list.append(
                _Bezier(
                    degrees=[1, 1, 1],
                    control_points=_np.array(
                        [
                            [y_max, v_one, v_one - y_max],
                            [v_one - y_max, v_one, v_one - y_max],
                            [v_zero, v_one, v_one],
                            [v_one, v_one, v_one],
                            [center, v_one - center, v_one - center],
                            [v_one - center, v_one - center, v_one - center],
                            [z_max, v_one - z_max, v_one],
                            [v_one - z_max, v_one - z_max, v_one],
                        ]
                    ),
                )
            )

            # x_max_y_min
            spline_list.append(
                _Bezier(
                    degrees=[1, 1, 1],
                    control_points=_np.array(
                        [
                            [v_one - center, center, center],
                            [v_one, x_max, x_max],
                            [v_one - center, center, v_one - center],
                            [v_one, x_max, v_one - x_max],
                            [v_one - y_min, v_zero, y_min],
                            [v_one, v_zero, v_zero],
                            [v_one - y_min, v_zero, v_one - y_min],
                            [v_one, v_zero, v_one],
                        ]
                    ),
                )
            )

            # x_min_y_min
            spline_list.append(
                _Bezier(
                    degrees=[1, 1, 1],
                    control_points=_np.array(
                        [
                            [v_zero, x_min, x_min],
                            [center, center, center],
                            [v_zero, x_min, v_one - x_min],
                            [center, center, v_one - center],
                            [v_zero, v_zero, v_zero],
                            [y_min, v_zero, y_min],
                            [v_zero, v_zero, v_one],
                            [y_min, v_zero, v_one - y_min],
                        ]
                    ),
                )
            )

            # y_min_z_max
            spline_list.append(
                _Bezier(
                    degrees=[1, 1, 1],
                    control_points=_np.array(
                        [
                            [center, center, v_one - center],
                            [v_one - center, center, v_one - center],
                            [z_max, z_max, v_one],
                            [v_one - z_max, z_max, v_one],
                            [y_min, v_zero, v_one - y_min],
                            [v_one - y_min, v_zero, v_one - y_min],
                            [v_zero, v_zero, v_one],
                            [v_one, v_zero, v_one],
                        ]
                    ),
                )
            )

            # y_min_z_min
            spline_list.append(
                _Bezier(
                    degrees=[1, 1, 1],
                    control_points=_np.array(
                        [
                            [z_min, z_min, v_zero],
                            [v_one - z_min, z_min, v_zero],
                            [center, center, center],
                            [v_one - center, center, center],
                            [v_zero, v_zero, v_zero],
                            [v_one, v_zero, v_zero],
                            [y_min, v_zero, y_min],
                            [v_one - y_min, v_zero, y_min],
                        ]
                    ),
                )
            )

            # Pass to output
            if i_derivative == 0:
                splines = spline_list.copy()
            else:
                derivatives.append(spline_list)

        # Return results
        return (splines, derivatives)



import numpy as np
import pytest

import splinepy


def test_volume_integration_1D(np_rng):
    """
    Test volume integration for splines using numerical integration of the
    Jacobi-Determinant
    """
    bezier = splinepy.Bezier(degrees=[1], control_points=[[0], [2]])

    assert np.allclose(bezier.integrate.volume(), 2.0)

    # test for other types same spline
    assert np.allclose(bezier.bspline.integrate.volume(), 2.0)
    assert np.allclose(bezier.rationalbezier.integrate.volume(), 2.0)
    assert np.allclose(bezier.nurbs.integrate.volume(), 2.0)

    # Check if equal after refinement
    bezier.elevate_degrees([0, 0, 0])

    assert np.allclose(bezier.integrate.volume(), 2.0)

    # Test knot insertion
    bspline = bezier.bspline
    bspline.insert_knots(0, np_rng.random(10))
    assert np.allclose(bezier.integrate.volume(), 2.0)


def test_volume_integration_embedded(np_rng):
    """
    Test volume integration for splines using numerical integration of the
    Jacobi-Determinant
    """
    # Test 1D -> 2D
    expected_result = 2.0**1.5
    bezier = splinepy.Bezier(degrees=[1], control_points=[[0, 0], [2, 2]])

    assert np.allclose(bezier.integrate.volume(), expected_result)

    # test for other types same spline
    assert np.allclose(bezier.bspline.integrate.volume(), expected_result)
    assert np.allclose(
        bezier.rationalbezier.integrate.volume(), expected_result
    )
    assert np.allclose(bezier.nurbs.integrate.volume(), expected_result)

    # Check if equal after refinement
    bezier.elevate_degrees([0, 0, 0])

    assert np.allclose(bezier.integrate.volume(), expected_result)

    # Test knot insertion
    bspline = bezier.bspline
    bspline.insert_knots(0, np_rng.random(10))
    assert np.allclose(bezier.integrate.volume(), expected_result)


def test_volume_integration_2D(np_rng):
    """
    Test volume integration for splines using numerical integration of the
    Jacobi-Determinant
    """
    # Test 2D
    bezier = splinepy.Bezier(
        degrees=[1, 1], control_points=[[0, 0], [2, 0], [0, 1], [2, 1]]
    )

    assert np.allclose(bezier.integrate.volume(), 2.0)

    # test for other types same spline
    assert np.allclose(bezier.bspline.integrate.volume(), 2.0)
    assert np.allclose(bezier.rationalbezier.integrate.volume(), 2.0)
    assert np.allclose(bezier.nurbs.integrate.volume(), 2.0)

    # Check if equal after refinement
    bezier.elevate_degrees([0, 0, 1])

    assert np.allclose(bezier.integrate.volume(), 2.0)

    # Test knot insertion
    bspline = bezier.bspline
    bspline.insert_knots(0, np_rng.random(10))
    assert np.allclose(bezier.integrate.volume(), 2.0)


def test_volume_integration_3D(np_rng):
    """
    Test volume integration for splines using numerical integration of the
    Jacobi-Determinant
    """
    # Test 3D
    bezier = splinepy.Bezier(
        degrees=[1, 1, 1],
        control_points=[
            [0, 0, 0],
            [1, 0, 0],
            [0, 1, 0],
            [1, 1, 0],
            [0, 0, 3],
            [1, 0, 3],
            [0, 1, 3],
            [1, 1, 3],
        ],
    )

    assert np.allclose(bezier.integrate.volume(), 3.0)

    # Check if equal after refinement
    bezier.elevate_degrees([0, 0, 1, 1, 2, 2])

    assert np.allclose(bezier.integrate.volume(), 3.0)

    # Test knot insertion
    bspline = bezier.bspline
    bspline.insert_knots(0, np_rng.random(10))
    assert np.allclose(bezier.integrate.volume(), 3.0)

    # Move control points along axis does not change volume
    mi = bspline.multi_index
    bspline.cps[mi[3, :, :]][:, 1] += 10
    assert np.allclose(bezier.integrate.volume(), 3.0)


def test_complex_geometry(np_rng):
    """Test on a more complex geometry"""
    bezier = splinepy.Bezier(
        degrees=[1, 1, 1],
        control_points=[
            [0, 0, 0],
            [1, 0, 0],
            [0, 1, 0],
            [1, 1, 0],
            [0, 0, 3],
            [1, 0, 3],
            [0, 1, 3],
            [1, 1, 3],
        ],
    )
    bezier.elevate_degrees([0, 1, 2, 0, 1, 2])
    bezier.cps += np_rng.random(bezier.cps.shape) * 0.1

    bezier_c = bezier.copy()

    # Move around internal control points

    mi = bezier.multi_index
    bezier.cps[mi[1:3, 1:3, 1:3]][:] += 0.05 * np_rng.random(
        bezier.cps[mi[1:3, 1:3, 1:3]][:].shape
    )
    assert np.allclose(bezier.integrate.volume(), bezier_c.integrate.volume())


def test_assertions(np_rng):
    """Test the assertions in volume function"""
    bezier = splinepy.Bezier(
        degrees=[1, 2], control_points=np_rng.random((6, 1))
    )

    with pytest.raises(
        ValueError, match=r"`Volume` not supported if para_dim > dim"
    ):
        bezier.integrate.volume()

    bezier = splinepy.Bezier(
        degrees=[2, 2], control_points=np_rng.random((9, 2))
    )

    with pytest.raises(
        ValueError, match=r"Integration order must be array of size para_dim"
    ):
        bezier.integrate.volume(orders=[2, 4, 5])


def test_function_integration(np_rng):
    col1_factor = 2

    def volume_function(x):
        vf = np.ones((x.shape[0], 2))
        # scale it with a factor to get a different value
        vf[:, 1] = col1_factor
        return vf

    bezier = splinepy.Bezier(
        degrees=[1, 2], control_points=np_rng.random((6, 2))
    )
    assert np.allclose(
        [bezier.integrate.volume(), col1_factor * bezier.integrate.volume()],
        bezier.integrate.parametric_function(volume_function),
    )



import numpy as np
import pytest

import splinepy

# frequently used fixtures
all_2p2d_splines = (
    "rational_bezier_2p2d",
    "bezier_2p2d",
    "bspline_2p2d",
    "nurbs_2p2d",
)

all_splines = (
    "rational_bezier_2p2d",
    "bezier_2p2d",
    "bspline_2p2d",
    "nurbs_2p2d",
    "rational_bezier_3p3d",
    "bezier_3p3d",
    "bspline_3p3d",
    "nurbs_3p3d",
)


@pytest.mark.parametrize(
    "spline, reference",
    [
        (
            "bspline_2p2d",
            [
                [
                    9.4128804e-01,
                    3.8615940e-02,
                    1.9602000e-04,
                    1.9015920e-02,
                    7.8012000e-04,
                    3.9600000e-06,
                    9.6040000e-05,
                    3.9400000e-06,
                    2.0000000e-08,
                ],
                [
                    2.4010000e-01,
                    9.8500000e-03,
                    5.0000000e-05,
                    4.8020000e-01,
                    1.9700000e-02,
                    1.0000000e-04,
                    2.4010000e-01,
                    9.8500000e-03,
                    5.0000000e-05,
                ],
                [
                    1.6200000e-02,
                    2.7540000e-01,
                    5.1840000e-01,
                    3.6000000e-03,
                    6.1200000e-02,
                    1.1520000e-01,
                    2.0000000e-04,
                    3.4000000e-03,
                    6.4000000e-03,
                ],
                [
                    7.2000000e-03,
                    5.0400000e-02,
                    3.2400000e-02,
                    3.3600000e-02,
                    2.3520000e-01,
                    1.5120000e-01,
                    3.9200000e-02,
                    2.7440000e-01,
                    1.7640000e-01,
                ],
                [
                    4.0000000e-06,
                    6.4000000e-05,
                    3.2000000e-05,
                    7.9200000e-04,
                    1.2672000e-02,
                    6.3360000e-03,
                    3.9204000e-02,
                    6.2726400e-01,
                    3.1363200e-01,
                ],
            ],
        ),
        (
            "nurbs_2p2d",
            [
                [
                    9.75958864e-01,
                    1.39415582e-02,
                    9.95774782e-05,
                    9.85817035e-03,
                    1.40823820e-04,
                    1.00583311e-06,
                ],
                [
                    4.92908517e-01,
                    7.04119101e-03,
                    5.02916557e-05,
                    4.92908517e-01,
                    7.04119101e-03,
                    5.02916557e-05,
                ],
                [
                    9.50089457e-03,
                    1.20926646e-01,
                    7.69572460e-01,
                    1.05565495e-03,
                    1.34362940e-02,
                    8.55080511e-02,
                ],
                [
                    1.32410262e-02,
                    7.49025551e-02,
                    2.11856419e-01,
                    3.08957277e-02,
                    1.74772629e-01,
                    4.94331644e-01,
                ],
                [
                    4.18891419e-03,
                    3.94934617e-03,
                    1.86173964e-03,
                    4.14702505e-01,
                    3.90985271e-01,
                    1.84312224e-01,
                ],
            ],
        ),
    ],
)
def test_basis_and_support(spline, reference, queries_2D, request):
    """Test the correct calculation of the basis functions.
    (.basis_and_support())"""
    spline = request.getfixturevalue(spline)
    # test basis functions
    assert np.allclose(
        spline.basis_and_support(queries_2D)[0],
        reference,
    )


def test_jacobian(np_rng):
    """Test the correct evaluation of basis function derivatives"""
    # Test both for nurbs as for rational beziers for the integration of a
    # circle, just because we can
    # Get Legendre points
    interval = [0, 1]
    order = 7
    dim = 2
    positions, weights = np.polynomial.legendre.leggauss(order)
    positions = interval[0] + (interval[1] - interval[0]) / 2 * (positions + 1)
    weights = weights * (interval[1] - interval[0]) / 2
    positions = np.reshape(
        np.meshgrid(*[positions for _ in range(dim)]), (dim, -1)
    ).T
    weights = np.prod(
        np.reshape(np.meshgrid(*[weights for _ in range(dim)]), (dim, -1)).T,
        axis=1,
    )
    inv_sqrt_2 = np.sqrt(0.5)
    circle_bezier = splinepy.RationalBezier(
        degrees=[2, 2],
        control_points=[
            [0, 0],
            [0.5, -0.5],
            [1, 0],
            [-0.5, 0.5],
            [0.5, 0.5],
            [1 + 0.5, 0.5],
            [0, 1],
            [0.5, 1 + 0.5],
            [1, 1],
        ],
        weights=[
            1,
            inv_sqrt_2,
            1,
            inv_sqrt_2,
            1,
            inv_sqrt_2,
            1,
            inv_sqrt_2,
            1,
        ],
    )
    circle_nurbs = splinepy.NURBS(
        degrees=circle_bezier.degrees,
        control_points=circle_bezier.control_points,
        knot_vectors=[[0, 0, 0, 1, 1, 1], [0, 0, 0, 1, 1, 1]],
        weights=circle_bezier.weights,
    )

    # Test the determinant
    area = np.sum(np.linalg.det(circle_bezier.jacobian(positions)) * weights)
    assert area == pytest.approx(np.pi * 0.5)

    area = np.sum(np.linalg.det(circle_nurbs.jacobian(positions)) * weights)
    assert area == pytest.approx(np.pi * 0.5)

    # Assuming that the derivative is correct
    test_spline = splinepy.Bezier(
        degrees=[3, 3], control_points=np_rng.random((16, 3))
    )
    query = np_rng.random((1, 2))
    jacs = test_spline.jacobian(query)[0]
    expected_jacs = np.vstack(
        (
            test_spline.derivative(query, [1, 0]),
            test_spline.derivative(query, [0, 1]),
        )
    ).T

    assert np.allclose(jacs, expected_jacs)

    # Test for scalar valued splines
    test_spline_scalar = splinepy.Bezier(
        degrees=[3, 3], control_points=np_rng.random((16, 1))
    )
    query = np_rng.random((1, 2))
    jacs = test_spline_scalar.jacobian(query)[0]
    expected_jacs = np.vstack(
        (
            test_spline_scalar.derivative(query, [1, 0]),
            test_spline_scalar.derivative(query, [0, 1]),
        )
    ).T

    assert np.allclose(jacs, expected_jacs)


def test_basis_function_derivatives(
    np_rng, bezier_2p2d, rational_bezier_2p2d, nurbs_2p2d
):
    """Test the correct evaluation of basis function derivatives"""
    # Cross-testing different libraries
    # use random query points
    q2D = np_rng.random((10, 2))

    # Rational Bezier and NURBS are equivalent but use different backends
    # increase orders for derivatives
    for _ in range(2):
        rational_bezier_2p2d.elevate_degrees([0, 1])
        nurbs_2p2d.elevate_degrees([0, 1])

    # Test different derivatives - all have global supports
    assert np.allclose(
        rational_bezier_2p2d.basis_derivative_and_support(q2D, [1, 0])[0],
        nurbs_2p2d.basis_derivative_and_support(q2D, [1, 0])[0],
    )

    assert np.allclose(
        rational_bezier_2p2d.basis_derivative_and_support(q2D, [1, 0])[1],
        nurbs_2p2d.basis_derivative_and_support(q2D, [1, 0])[1],
    )

    assert np.allclose(
        rational_bezier_2p2d.basis_derivative_and_support(q2D, [3, 2])[0],
        nurbs_2p2d.basis_derivative_and_support(q2D, [3, 2])[0],
    )

    assert np.allclose(
        rational_bezier_2p2d.basis_derivative_and_support(q2D, [1, 3])[0],
        nurbs_2p2d.basis_derivative_and_support(q2D, [1, 3])[0],
    )

    # For polynomial splines
    bspline_c = splinepy.BSpline(
        **bezier_2p2d.todict(),
        knot_vectors=[
            [0] * (bezier_2p2d.degrees[i] + 1)
            + [1] * (bezier_2p2d.degrees[i] + 1)
            for i in range(bezier_2p2d.para_dim)
        ],
    )

    # increase orders for derivatives
    for _i in range(2):
        bezier_2p2d.elevate_degrees([0, 1])
        bspline_c.elevate_degrees([0, 1])

    # Test different derivatives - all have global supports
    assert np.allclose(
        bezier_2p2d.basis_derivative_and_support(q2D, [0, 1])[0],
        bspline_c.basis_derivative_and_support(q2D, [0, 1])[0],
    )

    assert np.allclose(
        bezier_2p2d.basis_derivative_and_support(q2D, [0, 1])[1],
        bspline_c.basis_derivative_and_support(q2D, [0, 1])[1],
    )

    assert np.allclose(
        bezier_2p2d.basis_derivative_and_support(q2D, [2, 3])[0],
        bspline_c.basis_derivative_and_support(q2D, [2, 3])[0],
    )

    assert np.allclose(
        bezier_2p2d.basis_derivative_and_support(q2D, [3, 1])[0],
        bspline_c.basis_derivative_and_support(q2D, [3, 1])[0],
    )


def test_partition_of_unity(np_rng, bspline_2p2d, nurbs_2p2d):
    """Test the partition of unity of the calculated basis functions."""

    def basis_functions_sum(basis_functions):
        return basis_functions.sum(axis=1)

    # use random query points
    q2D = np_rng.random((10, 2))

    u_bspline = basis_functions_sum(bspline_2p2d.basis_and_support(q2D)[0])
    assert np.allclose(u_bspline, np.ones(np.shape(u_bspline)))

    u_nurbs = basis_functions_sum(nurbs_2p2d.basis_and_support(q2D)[0])
    assert np.allclose(u_nurbs, np.ones(np.shape(u_nurbs)))


@pytest.mark.parametrize(
    "spline, reference",
    [
        (
            "bspline_2p2d",
            [
                [-0.019796, 0.04049403],
                [-0.9996, 0.069675],
                [2.256, 1.9691],
                [1.528, 4.0766],
                [-1.0264, 2.873488],
            ],
        ),
        (
            "nurbs_2p2d",
            [
                [-1.00989841, 0.01432479],
                [-1.49984913, 0.02127445],
                [-0.15941144, 1.0883878],
                [-0.49948029, 1.62496752],
                [-1.61951381, 1.15640608],
            ],
        ),
        (
            "bezier_2p2d",
            [
                [-1.009899, 0.020099],
                [-1.49985, 0.02985],
                [-0.209, 1.089],
                [-0.612, 1.632],
                [-1.6716, 1.2736],
            ],
        ),
        (
            "rational_bezier_2p2d",
            [
                [-1.00989841, 0.01432479],
                [-1.49984913, 0.02127445],
                [-0.15941144, 1.0883878],
                [-0.49948029, 1.62496752],
                [-1.61951381, 1.15640608],
            ],
        ),
    ],
)
def test_evaluate(spline, reference, queries_2D, request):
    """Test the correct spline evaluation in the physical space.
    (.evaluate())"""
    spline = request.getfixturevalue(spline)
    # test evaluation
    assert np.allclose(
        spline.evaluate(queries_2D),
        reference,
    )


@pytest.mark.parametrize(
    "spline, reference",
    [
        (
            "bspline_2p2d",
            [
                [0.08, 7.8812],
                [0.08, 4.02],
                [0.8, 1.16],
                [1.6, 1.84],
                [3.2, 3.232],
            ],
        ),
        (
            "nurbs_2p2d",
            [
                [0.02017474, 1.42231975],
                [0.02017474, 1.42231975],
                [1.47716145, 0.21635343],
                [1.49159582, 0.45848468],
                [0.95624956, 1.33920031],
            ],
        ),
    ],
)
def test_derivative(spline, reference, queries_2D, request):
    """Test the correct calculation of the first derivative.
    (.derivative())"""
    spline = request.getfixturevalue(spline)

    # order
    o1 = [1, 1]

    # test derivative evaluation
    assert np.allclose(
        spline.derivative(queries_2D, o1),
        reference,
    )


def test_higher_order_derivatives(np_rng):
    """Tests higher order derivatives of the spline functions."""
    # Test higher order derivatives against each other
    max_deg = 5
    dim = 3
    n_test_points = 10
    derivatives = np_rng.integers(0, max_deg, size=(dim))
    queries = np_rng.random((n_test_points, dim))
    randomized = {
        "degrees": [5] * dim,
        "control_points": np_rng.random(((max_deg + 1) ** dim, dim)),
        "weights": np_rng.random((max_deg + 1) ** dim),
        "knot_vectors": [[0] * (max_deg + 1) + [1] * (max_deg + 1)] * dim,
    }
    random_bezier = splinepy.Bezier(
        degrees=randomized["degrees"],
        control_points=randomized["control_points"],
    )
    random_rational = splinepy.RationalBezier(
        degrees=randomized["degrees"],
        control_points=randomized["control_points"],
        weights=randomized["weights"],
    )
    random_bspline = splinepy.BSpline(
        degrees=randomized["degrees"],
        control_points=randomized["control_points"],
        knot_vectors=randomized["knot_vectors"],
    )
    random_nurbs = splinepy.NURBS(
        degrees=randomized["degrees"],
        control_points=randomized["control_points"],
        knot_vectors=randomized["knot_vectors"],
        weights=randomized["weights"],
    )
    assert np.allclose(
        random_bezier.derivative(queries, derivatives),
        random_bspline.derivative(queries, derivatives),
    )

    assert np.allclose(
        random_rational.derivative(queries, derivatives),
        random_nurbs.derivative(queries, derivatives),
    )


@pytest.mark.parametrize("splinetype", all_2p2d_splines)
def test_assertions_evaluation(request, splinetype):
    """Test if the right assertions are thrown"""
    spline = request.getfixturevalue(splinetype)
    # Test for evaluation
    # Check minimum with kwargs

    with pytest.raises(
        ValueError,
        match=r"Query request out of bounds in parametric dimension 0. "
        r"Detected query \[-0.1  0. \] at positions 0, which is out of"
        r" bounds with minimum values \[1. 1.\].",
    ):
        spline.evaluate(queries=[[-0.1, 0.0], [0.4, 0.6]])

    # Check maximum with args
    with pytest.raises(
        ValueError,
        match=r"Query request out of bounds in parametric dimension 0. "
        r"Detected query \[1.1 0. \] at positions 0, which is out of "
        r"bounds with maximum values \[1. 1.\].",
    ):
        spline.evaluate([[1.1, 0.0], [0.4, 0.6]])

    # Check dimensions
    with pytest.raises(
        ValueError,
        match=r"Dimension mismatch between parametric dimension of spline "
        r"\(2\), and query-request \(3\)",
    ):
        spline.evaluate([[1.0, 1.0, 1.0]])


@pytest.mark.parametrize("splinetype", all_2p2d_splines)
def test_assertions_jacobian(request, splinetype):
    """Test if the right assertions are thrown"""
    # Test for evaluation
    spline = request.getfixturevalue(splinetype)
    # Check minimum with kwargs
    with pytest.raises(
        ValueError,
        match=r"Query request out of bounds in parametric dimension 0. "
        r"Detected query \[-0.1  0. \] at positions 0, which is out of"
        r" bounds with minimum values \[1. 1.\].",
    ):
        spline.jacobian(queries=[[-0.1, 0.0], [0.4, 0.6]])

    # Check maximum with args
    with pytest.raises(
        ValueError,
        match=r"Query request out of bounds in parametric dimension 0. "
        r"Detected query \[1.1 0. \] at positions 0, which is out of "
        r"bounds with maximum values \[1. 1.\].",
    ):
        spline.jacobian([[1.1, 0.0], [0.4, 0.6]])

    # Check dimensions
    with pytest.raises(
        ValueError,
        match=r"Dimension mismatch between parametric dimension of spline "
        r"\(2\), and query-request \(3\)",
    ):
        spline.jacobian([[1.0, 1.0, 1.0]])


@pytest.mark.parametrize("splinetype", all_2p2d_splines)
def test_basis_function_matrix(request, splinetype, np_rng):
    """Test the correct evaluation of basis function matrices"""
    make_matrix = splinepy.utils.data.make_matrix
    # Use
    q2D = np_rng.random((10, 2))

    if splinetype == "nurbs_2p2d":
        return

    # Test for both bezier and BSpline family and rationals
    else:
        spline = request.getfixturevalue(splinetype)
        # Compute derivative matrix
        # Trivial evaluation
        matrix = make_matrix(
            *spline.basis_and_support(q2D), spline.cps.shape[0]
        )
        assert np.allclose(
            matrix @ spline.cps,
            spline.evaluate(q2D),
        )

        # Test first order derivative
        matrix = make_matrix(
            *spline.basis_derivative_and_support(q2D, orders=[1, 0]),
            spline.cps.shape[0],
        )
        assert np.allclose(
            matrix @ spline.cps,
            spline.derivative(q2D, orders=[1, 0]),
        )

        # Test seoncd order derivative as numpy (enforce)
        matrix = make_matrix(
            *spline.basis_derivative_and_support(q2D, orders=[1, 2]),
            spline.cps.shape[0],
            as_array=True,
        )
        assert np.allclose(
            matrix @ spline.cps,
            spline.derivative(q2D, orders=[1, 2]),
        )


@pytest.mark.parametrize("splinetype", all_splines)
def test_multiple_derivative_queries(request, splinetype, np_rng):
    """Test cartesian product queries of parametric coordinates and orders"""
    spline = request.getfixturevalue(splinetype)

    p_coord = {2: np_rng.random((10, 2)), 3: np_rng.random((10, 3))}
    a = np.array  # shortcut
    orders = {
        2: splinepy.utils.data.cartesian_product([a([0, 1, 2]), a([0, 1, 2])]),
        3: splinepy.utils.data.cartesian_product(
            [a([0, 1, 2]), a([0, 1, 2]), a([0, 1, 2])]
        ),
    }
    jac_transpose_orders = {2: np.eye(2), 3: np.eye(3)}

    pd = spline.para_dim
    qs = p_coord[pd]
    jto = jac_transpose_orders[pd]
    os = orders[pd]

    # multiple queries
    multi0 = spline.derivative(qs, os)
    multi1 = spline.basis_derivative(qs, os)
    multi2, multi2_support = spline.basis_derivative_and_support(qs, os)

    # as well as jacobian transposed queries
    # however, we transpose results to enable direct comparison
    multi_jac = spline.derivative(qs, jto).transpose(0, 2, 1)
    single_jac = spline.jacobian(qs)

    # multiple single evaluations
    single0 = []
    single1 = []
    single2 = []
    single2_support = []
    # loop in order of the visit
    for q in qs:
        for o in os:
            single0.append(spline.derivative([q], o))
            single1.append(spline.basis_derivative([q], o))
            s2, s2_s = spline.basis_derivative_and_support([q], o)
            single2.append(s2)
            single2_support.append(s2_s)

    # compare and see if they match
    # multi queries will return
    # (n_queries, n_orders, dim or n_support) shape
    # for comparison, reshape. we could also ravel both, alternatively.
    dim = multi0.shape[-1]
    assert np.allclose(multi0.reshape(-1, dim), np.vstack(single0))
    sup = multi1.shape[-1]
    assert np.allclose(multi1.reshape(-1, sup), np.vstack(single1))
    assert np.allclose(multi2.reshape(-1, sup), np.vstack(single2))
    assert (
        multi2_support
        == a(single2_support).reshape(-1, multi2_support.shape[-1])[:: len(os)]
    ).all()
    assert np.allclose(multi_jac, single_jac)



import numpy as np
import pytest

import splinepy

# fixtures for test_new_control_point_pointers_creation
all_2p2d_splines = (
    "rational_bezier_2p2d",
    "bezier_2p2d",
    "bspline_2p2d",
    "nurbs_2p2d",
)


# often called function; no testfunction
def cps_are_synced(spl):
    assert np.allclose(
        spl.cps, spl.current_core_properties()["control_points"]
    )


@pytest.mark.parametrize("splinetype", all_2p2d_splines)
def test_inplace_change_degrees_para(splinetype, request):
    """inplace change of degrees should not be allowed if core spline is
    initialized"""
    spline = request.getfixturevalue(splinetype)

    spline_dict = spline.todict()
    s = spline.__class__()
    s.degrees = spline_dict["degrees"]
    s.degrees += 1

    # this shoundn't be fine
    s._new_core(**spline_dict)
    with pytest.raises(ValueError):
        s.degrees += 1


def test_inplace_change_knot_vectors(nd_box, raster):
    """test inplace change of knot_vectors"""
    # let's test 3D splines
    dim = 3
    box_data = nd_box(dim)
    n = splinepy.NURBS(**box_data)
    box_data.pop("weights")
    b = splinepy.BSpline(**box_data)

    # elevate degrees and insert some knots
    knot_insert_dims = [1, 2]
    for s in (n, b):
        s.elevate_degrees([0, 0, 1, 2])
        for kid in knot_insert_dims:
            s.insert_knots(kid, np.linspace(0.1, 0.9, 9))

    qres = 5
    raster_query = raster([[0] * dim, [1] * dim], [qres] * dim)

    for s in (n, b):
        # is this valid spline?
        assert np.allclose(raster_query, s.evaluate(raster_query))

        # modify knots and corresponding queries, so that
        # evaluated points are same as raster_query
        modified_query = raster_query.copy()
        factor = 2
        for kid in knot_insert_dims:
            s.knot_vectors[kid][:] = np.multiply(
                s.knot_vectors[kid], factor
            ).tolist()
            modified_query[:, kid] *= factor

        # evaluation check
        assert np.allclose(raster_query, s.evaluate(modified_query))


def test_inplace_change_control_points(nd_box):
    """test inplace changes of control points"""
    dim = 3
    res = [3] * dim
    box_data = nd_box(dim)

    n = splinepy.NURBS(**box_data)
    weights = box_data.pop("weights")
    b = splinepy.BSpline(**box_data)
    box_data.pop("knot_vectors")
    z = splinepy.Bezier(**box_data)
    r = splinepy.RationalBezier(**box_data, weights=weights)

    for s in (z, r, b, n):
        # init
        orig = s.copy()

        # check if we are good to start
        assert np.allclose(
            orig.sample(res), s.sample(res)
        ), f"{s.whatami} failed"

        # hold reference to control points
        s_cps = s.control_points
        # modify cps
        s.control_points /= 2
        assert np.allclose(
            orig.sample(res) / 2, s.sample(res)
        ), f"{s.whatami} failed"
        # check if they are still the same array
        assert s_cps is s.control_points, f"{s.whatami} failed"

        # modify ws
        if s.is_rational:
            s_ws = s.weights
            s.weights *= 0.5
            assert s_ws is s.weights, f"{s.whatami} failed"


def test_inplace_change_weights(nurbs_2p2d):
    """test inplace change of weights by comparing quarter circle"""
    n_q_circle = nurbs_2p2d.todict()
    res = [4] * 2
    # init rational
    n = splinepy.NURBS(**n_q_circle)
    kvs = n_q_circle.pop("knot_vectors")
    r = splinepy.RationalBezier(**n_q_circle)
    # init non-rational
    n_q_circle.pop("weights")
    b = splinepy.BSpline(**n_q_circle, knot_vectors=kvs)
    z = splinepy.Bezier(**n_q_circle)

    # make sure they aren't the same
    assert not np.allclose(b.sample(res), n.sample(res))
    assert not np.allclose(z.sample(res), r.sample(res))

    # modify weights of rational splines
    for rs in (n, r):
        # set them all to 1
        rs.weights[abs(rs.weights - 1.0) > 1e-10] = 1.0

    # now, they should be the same
    assert np.allclose(b.sample(res), n.sample(res))
    assert np.allclose(z.sample(res), r.sample(res))
    assert np.allclose(b.sample(res), r.sample(res))
    assert np.allclose(z.sample(res), n.sample(res))


def test_physical_space_array(nurbs_2p2d):
    """tests if physical space array is syncing correctly."""
    n = splinepy.NURBS(**nurbs_2p2d.todict())

    # all setitem cases
    # 1.
    n.cps[0] = 100.0
    cps_are_synced(n)

    n.cps[np.array(1)] = [0.99, 0.11]  # here, weight is not 1
    cps_are_synced(n)

    # 2.
    n.cps[np.array([1, 4, 3])] = 123
    cps_are_synced(n)

    b_mask = np.ones(len(n.cps), dtype=bool)
    b_mask[[0, 4]] = False
    n.cps[b_mask] = -1235.5
    cps_are_synced(n)

    # 3.
    n.cps[[2, 0, 5]] = 299354.0
    cps_are_synced(n)

    n.cps[b_mask.tolist()] = 975.5
    cps_are_synced(n)

    # 4.
    n.cps[3, :2] = [-10, -20]
    cps_are_synced(n)

    n.cps[5, [1, 0]] = [1.722, 24.29]
    cps_are_synced(n)

    n.cps[np.array(2), 1] = 7
    cps_are_synced(n)

    # 5.
    n.cps[[0, 5, 2], 0] = 92314.111234
    cps_are_synced(n)
    n.cps[b_mask, :2] = n.cps[[2, 5, 0, 1], -2:]
    cps_are_synced(n)
    n.cps[b_mask.tolist(), 0] = n.cps[b_mask[::-1], 1]
    cps_are_synced(n)

    # 6.
    n.cps[:1, :2] = 160.4
    cps_are_synced(n)
    n.cps[2:-1, :2] = 78883.2
    n.cps[-1:-4:-1, :1] = -1687
    cps_are_synced(n)

    # 7.
    n.cps[...] = 1237792.4634527
    n.cps[..., :1] = -62.5
    cps_are_synced(n)

    # 8. ufunc
    np.add(n.cps, n.cps, out=n.cps)
    cps_are_synced(n)
    np.multiply(n.cps, n.cps, out=n.cps)
    cps_are_synced(n)

    # 9. if you have ideas for use case not listed above, please add!

    # now, child arrays
    carr = n.cps[0]
    assert carr.base is n.cps

    carr += [1324235.1, -234.5]
    assert np.allclose(carr, n.cps[0])
    cps_are_synced(n)

    # copy should return a normal np.ndarray
    ndarr = n.cps.copy()
    assert isinstance(ndarr, np.ndarray)
    assert not isinstance(ndarr, type(n.cps))
    assert not isinstance(ndarr, splinepy.utils.data.PhysicalSpaceArray)


@pytest.mark.parametrize("splinetype", all_2p2d_splines)
def test_new_control_point_pointers_creation(splinetype, request):
    """Test cp pointers are newly created if they change.
    _sync_source_ptr compares len of cp pointers and PhysicalSpaceArray
    before syncing.
    """
    # get spline
    spline = request.getfixturevalue(splinetype)

    # initial check
    spline.cps._sync_source_ptr()
    cps_are_synced(spline)

    # check after degree elevation
    spline.elevate_degrees([0, 1])
    spline.cps._sync_source_ptr()
    cps_are_synced(spline)

    # and after knot vectors and reduce degree
    if spline.has_knot_vectors:
        spline.insert_knots(0, [0.2, 0.5, 0.7])
        spline.cps._sync_source_ptr()
        cps_are_synced(spline)

        spline.insert_knots(1, [0.2, 0.5, 0.7])
        spline.cps._sync_source_ptr()
        cps_are_synced(spline)

        spline.remove_knots(0, [0.7])
        spline.cps._sync_source_ptr()
        cps_are_synced(spline)

        # reduce degree only available for bsplines
        spline.reduce_degrees([0])
        spline.cps._sync_source_ptr()
        cps_are_synced(spline)



"""
A script for extra doc generation.
Feel free to extend!

Contents:
1. Create markdown table of show options.
"""

import os

import splinepy as spp

if __name__ == "__main__":
    here = os.path.dirname(os.path.abspath(__file__))

    # create md dir
    md_path = os.path.abspath(os.path.join(here, "..", "md"))
    os.makedirs(md_path, exist_ok=True)

    # 1. append show options to visualization
    with open(
        os.path.abspath(os.path.join(here, "../md/spline_plotting.md")), "a"
    ) as f:
        f.write("# List of show_options\n")
        derived = [
            spp.helpme.visualize.SplineShowOption,
            spp.helpme.visualize.MultipatchShowOption,
        ]
        for cls in derived:
            f.write(f"## {cls.__qualname__}\n\n")
            for option in cls._valid_options.values():
                t_str = str(option.allowed_types)
                t_str = (
                    t_str.replace("<class", "")
                    .replace("'", "")
                    .replace(">", "")
                )
                f.write(
                    f"<details><summary><strong>{option.key}"
                    "</strong></summary><p>\n"
                )
                f.write(f"\n{option.description}  \n")
                f.write(f"- _allowed types_: {t_str}  \n")
                f.write(f"- _default_: {option.default}  \n")
                f.write("</p></details> \n\n")



"""splinepy/splinepy/log.py
Thin logging wrapper.
"""

import functools as _functools
import logging as _logging


def configure(debug=False, logfile=None):
    """
    Logging configurator. Can help you to set debug or info.
    Calling this will keep only one logging.StreamHandler.

    Parameters
    -----------
    debug: bool
    logfile: str

    Returns
    --------
    None
    """
    # logger
    logger = _logging.getLogger("splinepy")

    # level
    level = _logging.DEBUG if debug else _logging.INFO
    logger.setLevel(level)

    # format
    formatter = _logging.Formatter(
        fmt="%(asctime)-15s %(name)s [%(levelname)s] %(message)s"
    )

    # apply format using stream handler
    # let's use only one stream handler so that calling configure multiple
    # times won't duplicate printing.
    new_handlers = []
    for _i, h in enumerate(logger.handlers):
        # we skip all the stream handler.
        if isinstance(h, _logging.StreamHandler):
            continue

        # blindly keep other ones
        else:
            new_handlers.append(h)

    # add new stream handler
    stream_handler = _logging.StreamHandler()
    stream_handler.setLevel(level)
    stream_handler.setFormatter(formatter)
    new_handlers.append(stream_handler)

    logger.handlers = new_handlers

    # output logs
    if logfile is not None:
        file_logger_handler = _logging.FileHandler(logfile)
        logger.addHandler(file_logger_handler)


def debug(*log):
    """
    Debug logger.

    Parameters
    -----------
    *log: str

    Returns
    --------
    None
    """
    logger = _logging.getLogger("splinepy")
    logger.debug(" ".join(map(str, log)))


def info(*log):
    """
    Info logger.

    Parameters
    -----------
    *log: str

    Returns
    --------
    None
    """
    logger = _logging.getLogger("splinepy")
    logger.info(" ".join(map(str, log)))


def warning(*log):
    """
    warning logger.

    Parameters
    -----------
    *log: str

    Returns
    --------
    None
    """
    logger = _logging.getLogger("splinepy")
    logger.warning(" ".join(map(str, log)))


def error(*log):
    """
    error logger.

    Parameters
    -----------
    *log: str

    Returns
    --------
    None
    """
    logger = _logging.getLogger("splinepy")
    logger.error(" ".join(map(str, log)))


def prepend_log(message, log_func):
    """
    Prepend message before a logging function.

    Parameters
    ----------
    message: str
    log_func: function
      one of the following - {info, debug, warning}

    Returns
    -------
    prepended: function
    """
    return _functools.partial(log_func, message)



from gustaf import show  # shortcut for frequently used module / func

from splinepy import (
    bezier,
    bspline,
    helpme,
    io,
    microstructure,
    multipatch,
    nurbs,
    rational_bezier,
    settings,
    spline,
    splinepy_core,
    utils,
)
from splinepy._version import __version__
from splinepy.bezier import Bezier
from splinepy.bspline import BSpline
from splinepy.helpme.ffd import FFD
from splinepy.microstructure.microstructure import Microstructure
from splinepy.multipatch import Multipatch
from splinepy.nurbs import NURBS
from splinepy.rational_bezier import RationalBezier
from splinepy.spline import Spline
from splinepy.utils.data import (
    SplineDataAdaptor,
    cartesian_product,
    make_matrix,
    uniform_query,
)

# set NAME_TO_TYPE
settings.NAME_TO_TYPE = settings.__splinepy_name_to_type__()

# configure logging
utils.log.configure()


def to_derived(spline):
    """
    Returns derived spline type based on NAME_TO_TYPE conversion.

    Parameters
    ----------
    spline: CoreSpline

    Returns
    -------
    derived_spline: DerivedSpline
    """
    return settings.NAME_TO_TYPE[spline.name](spline=spline)


__all__ = [
    "__version__",
    "bezier",
    "bspline",
    "io",
    "multipatch",
    "nurbs",
    "rational_bezier",
    "spline",
    "splinepy_core",
    "utils",
    "microstructure",
    "settings",
    "helpme",
    "Bezier",
    "BSpline",
    "Multipatch",
    "NURBS",
    "RationalBezier",
    "Spline",
    "SplineDataAdaptor",
    "FFD",
    "to_derived",
    "show",
    "Microstructure",
    "TileLib",
    "cartesian_product",
    "make_matrix",
    "uniform_query",
]
